import shutil
from contextlib import contextmanager

import pytest

from indy_common.config_helper import NodeConfigHelper
from indy_node.test.helper import TestNode
from plenum.test.node_catchup.helper import ensure_all_nodes_have_same_data
from plenum.test.test_node import ensureElectionsDone, ensure_node_disconnected, checkNodesConnected
from indy_node.test.auth_rule.helper import sdk_send_and_check_auth_rule_request, sdk_send_and_check_get_auth_rule_request
from indy_common.authorize.auth_actions import ADD_PREFIX, AuthActionAdd
from indy_common.authorize.auth_constraints import AuthConstraint, ROLE
from indy_common.constants import CONSTRAINT, AUTH_TYPE, CONFIG_LEDGER_ID, NYM
from indy_common.authorize.auth_map import one_trustee_constraint
from plenum.common.constants import STEWARD, DATA
from plenum.common.exceptions import RequestNackedException


@contextmanager
def extend_auth_map(nodes, key, constraint):
    """
    Context manager to add a new auth rule to the auth map and remove it on exit.

    :param nodes: nodes list which auth maps should be changed
    :param key: str gotten from AuthActionAdd(...).get_action_id()
    :param constraint: AuthConstraint
    """
    for node in nodes:
        node.write_req_validator.auth_map[key] = constraint
    yield
    for node in nodes:
        node.write_req_validator.auth_map.pop(key, None)


def test_auth_txn_with_deprecated_key(tconf, tdir, allPluginsPath,
                                      txnPoolNodeSet,
                   looper,
                   sdk_wallet_trustee,
                   sdk_pool_handle):
    """
    Add to the auth_map a fake rule
    Send AUTH_RULE txn to change this fake rule (and set the fake key to the config state)
    Send GET_AUTH_RULE txn and check that the fake rule was changed
    Remove the fake auth rule from the map
    Check that we can't get the fake auth rule
    Restart the last node with its state regeneration
    Check that nodes data is equal after changing the existing auth rule (restarted node regenerate config state)
    """

    fake_txn_type = "100002"
    fake_key = AuthActionAdd(txn_type=fake_txn_type,
                                field="*",
                                value="*").get_action_id()
    fake_constraint = one_trustee_constraint
    new_auth_constraint = AuthConstraint(role=STEWARD, sig_count=1, need_to_be_owner=False).as_dict

    # Add to the auth_map a fake rule
    with extend_auth_map(txnPoolNodeSet,
                         fake_key,
                         fake_constraint):
        # Send AUTH_RULE txn to change this fake rule (and set the fake key to the config state)
        sdk_send_and_check_auth_rule_request(looper,
                                             sdk_pool_handle,
                                             sdk_wallet_trustee,
                                             auth_action=ADD_PREFIX,
                                             auth_type=fake_txn_type,
                                             field='*',
                                             new_value='*',
                                             constraint=new_auth_constraint)
        # Send GET_AUTH_RULE txn and check that the fake rule was changed
        result = sdk_send_and_check_get_auth_rule_request(
            looper,
            sdk_pool_handle,
            sdk_wallet_trustee,
            auth_type=fake_txn_type,
            auth_action=ADD_PREFIX,
            field="*",
            new_value="*"
        )[0][1]["result"][DATA][0]
        assert result[AUTH_TYPE] == fake_txn_type
        assert result[CONSTRAINT] == new_auth_constraint

    # Remove the fake auth rule from the map
    # Check that we can't get the fake auth rule
    with pytest.raises(RequestNackedException, match="not found in authorization map"):
        sdk_send_and_check_auth_rule_request(looper,
                                         sdk_pool_handle,
                                         sdk_wallet_trustee,
                                         auth_action=ADD_PREFIX,
                                         auth_type=fake_txn_type,
                                         field='*',
                                         new_value='*',
                                         constraint=AuthConstraint(role=STEWARD, sig_count=2,
                                                                   need_to_be_owner=False).as_dict)

    resp = sdk_send_and_check_get_auth_rule_request(looper,
                                                    sdk_pool_handle,
                                                    sdk_wallet_trustee)

    assert all(rule[AUTH_TYPE] != fake_txn_type for rule in resp[0][1]["result"][DATA])

    with pytest.raises(RequestNackedException, match="not found in authorization map"):
        sdk_send_and_check_get_auth_rule_request(
            looper,
            sdk_pool_handle,
            sdk_wallet_trustee,
            auth_type=fake_txn_type,
            auth_action=ADD_PREFIX,
            field="*",
            new_value="*"
        )
    # Restart the last node with its state regeneration
    ensure_all_nodes_have_same_data(looper, txnPoolNodeSet)

    node_to_stop = txnPoolNodeSet[-1]
    node_state = node_to_stop.states[CONFIG_LEDGER_ID]
    assert not node_state.isEmpty
    state_db_path = node_state._kv.db_path
    node_to_stop.cleanupOnStopping = False
    node_to_stop.stop()
    looper.removeProdable(node_to_stop)
    ensure_node_disconnected(looper, node_to_stop, txnPoolNodeSet[:-1])

    shutil.rmtree(state_db_path)

    config_helper = NodeConfigHelper(node_to_stop.name, tconf, chroot=tdir)
    restarted_node = TestNode(
        node_to_stop.name,
        config_helper=config_helper,
        config=tconf,
        pluginPaths=allPluginsPath,
        ha=node_to_stop.nodestack.ha,
        cliha=node_to_stop.clientstack.ha)
    looper.add(restarted_node)
    txnPoolNodeSet[-1] = restarted_node

    # Check that nodes data is equal (restarted node regenerate config state)
    looper.run(checkNodesConnected(txnPoolNodeSet))
    ensureElectionsDone(looper, txnPoolNodeSet, customTimeout=30)
    sdk_send_and_check_auth_rule_request(looper,
                                         sdk_pool_handle,
                                         sdk_wallet_trustee,
                                         auth_action=ADD_PREFIX,
                                         auth_type=NYM,
                                         field=ROLE,
                                         new_value=STEWARD,
                                         constraint=AuthConstraint(role=STEWARD, sig_count=2,
                                                                   need_to_be_owner=False).as_dict)
    ensure_all_nodes_have_same_data(looper, txnPoolNodeSet, custom_timeout=20)




import json
import pytest

from indy_common.constants import REVOKED, VALUE, PREV_ACCUM, CRED_DEF_ID, CLAIM_DEF_SCHEMA_REF, \
    CLAIM_DEF_SIGNATURE_TYPE, CLAIM_DEF_TAG, ACCUM
from indy_common.state.domain import make_state_path_for_claim_def
from indy_node.test.anon_creds.conftest import claim_def, build_revoc_reg_entry_for_given_revoc_reg_def, \
    build_revoc_def_by_default, build_revoc_def_by_endorser, build_revoc_def_by_steward, build_revoc_def_by_demand, \
    build_revoc_def_random
from indy_node.test.schema.test_send_get_schema import send_schema_seq_no
from plenum.common.exceptions import RequestNackedException, RequestRejectedException
from plenum.common.util import randomString
from plenum.test.helper import sdk_sign_request_from_dict, sdk_send_and_check

from indy_node.test.schema.test_send_get_schema import send_schema_req


def create_revoc_reg_def(looper, txnPoolNodeSet, sdk_pool_handle, build_revoc,
                         claim_def, wallet):
    # We need to have claim_def to send revocation txns

    claim_def_req = sdk_sign_request_from_dict(looper, wallet, claim_def)
    sdk_send_and_check([json.dumps(claim_def_req)], looper, txnPoolNodeSet, sdk_pool_handle)

    _, author_did = wallet
    revoc_reg = build_revoc
    revoc_reg['operation'][CRED_DEF_ID] = \
        make_state_path_for_claim_def(author_did,
                                      str(claim_def_req['operation'][CLAIM_DEF_SCHEMA_REF]),
                                      claim_def_req['operation'][CLAIM_DEF_SIGNATURE_TYPE],
                                      claim_def_req['operation'][CLAIM_DEF_TAG]
                                      ).decode()
    revoc_req = sdk_sign_request_from_dict(looper, wallet, revoc_reg['operation'])
    _, revoc_reply = sdk_send_and_check([json.dumps(revoc_req)], looper, txnPoolNodeSet, sdk_pool_handle)[0]
    return revoc_req


def create_revoc_reg_entry(looper, txnPoolNodeSet, sdk_pool_handle, build_revoc, claim_def, wallet):
    revoc_def_req = create_revoc_reg_def(looper, txnPoolNodeSet, sdk_pool_handle, build_revoc,
                                         claim_def, wallet)

    rev_reg_entry = build_revoc_reg_entry_for_given_revoc_reg_def(revoc_def_req)
    rev_reg_entry[VALUE][REVOKED] = [1, 2, 3, 4, 5]
    del rev_reg_entry[VALUE][PREV_ACCUM]
    rev_entry_req = sdk_sign_request_from_dict(looper, wallet, rev_reg_entry)
    sdk_send_and_check([json.dumps(rev_entry_req)], looper, txnPoolNodeSet, sdk_pool_handle)
    return rev_entry_req


def test_client_cant_create_revoc_reg_def(looper,
                                          txnPoolNodeSet,
                                          sdk_wallet_client,
                                          sdk_pool_handle,
                                          build_revoc_def_by_default,
                                          claim_def, tconf):
    with pytest.raises(RequestRejectedException):
        create_revoc_reg_def(looper, txnPoolNodeSet, sdk_pool_handle, build_revoc_def_by_default,
                             claim_def, sdk_wallet_client)


def test_allowed_roles_can_create_revoc_reg_def(looper,
                                                txnPoolNodeSet,
                                                sdk_wallet_trustee,
                                                sdk_wallet_endorser,
                                                sdk_wallet_steward,
                                                sdk_pool_handle,
                                                build_revoc_def_by_default,
                                                claim_def, tconf):
    # endorser
    create_revoc_reg_def(looper, txnPoolNodeSet, sdk_pool_handle, build_revoc_def_by_default,
                         claim_def, sdk_wallet_endorser)
    # steward
    create_revoc_reg_def(looper, txnPoolNodeSet, sdk_pool_handle, build_revoc_def_by_default,
                         claim_def, sdk_wallet_steward)
    # trustee
    create_revoc_reg_def(looper, txnPoolNodeSet, sdk_pool_handle, build_revoc_def_by_default,
                         claim_def, sdk_wallet_trustee)


def test_allowed_roles_can_send_revoc_reg_entry(looper,
                                                txnPoolNodeSet,
                                                sdk_wallet_trustee,
                                                sdk_wallet_endorser,
                                                sdk_wallet_steward,
                                                sdk_pool_handle,
                                                build_revoc_def_by_default,
                                                claim_def, tconf):
    # endorser
    create_revoc_reg_entry(looper, txnPoolNodeSet, sdk_pool_handle,
                           build_revoc_def_by_default, claim_def, sdk_wallet_endorser)

    # steward

    create_revoc_reg_entry(looper, txnPoolNodeSet, sdk_pool_handle,
                           build_revoc_def_by_default, claim_def, sdk_wallet_steward)

    # trustee
    create_revoc_reg_def(looper, txnPoolNodeSet, sdk_pool_handle,
                         build_revoc_def_by_default, claim_def, sdk_wallet_trustee)


def test_client_not_owner_cant_create_revoc_reg_entry(looper,
                                                      txnPoolNodeSet,
                                                      sdk_wallet_trustee,
                                                      sdk_wallet_client,
                                                      sdk_pool_handle,
                                                      build_revoc_def_by_default,
                                                      claim_def, tconf):
    revoc_def_req_trustee = create_revoc_reg_def(looper, txnPoolNodeSet, sdk_pool_handle, build_revoc_def_by_default,
                                                 claim_def, sdk_wallet_trustee)

    rev_reg_entry = build_revoc_reg_entry_for_given_revoc_reg_def(revoc_def_req_trustee)
    rev_reg_entry[VALUE][REVOKED] = [1, 2, 3, 4, 5]
    del rev_reg_entry[VALUE][PREV_ACCUM]
    rev_entry_req_client = sdk_sign_request_from_dict(looper, sdk_wallet_client, rev_reg_entry)
    with pytest.raises(RequestRejectedException):
        sdk_send_and_check([json.dumps(rev_entry_req_client)], looper, txnPoolNodeSet, sdk_pool_handle)


def test_endorser_not_owner_cant_create_revoc_reg_entry(looper,
                                                            txnPoolNodeSet,
                                                            sdk_wallet_trustee,
                                                            sdk_wallet_endorser,
                                                            sdk_pool_handle,
                                                            build_revoc_def_by_default,
                                                            claim_def, tconf):
    revoc_def_req_trustee = create_revoc_reg_def(looper, txnPoolNodeSet, sdk_pool_handle, build_revoc_def_by_default,
                                                 claim_def, sdk_wallet_trustee)

    rev_reg_entry = build_revoc_reg_entry_for_given_revoc_reg_def(revoc_def_req_trustee)
    rev_reg_entry[VALUE][REVOKED] = [1, 2, 3, 4, 5]
    del rev_reg_entry[VALUE][PREV_ACCUM]
    rev_entry_req_endorser = sdk_sign_request_from_dict(looper, sdk_wallet_endorser, rev_reg_entry)
    with pytest.raises(RequestRejectedException):
        sdk_send_and_check([json.dumps(rev_entry_req_endorser)], looper, txnPoolNodeSet, sdk_pool_handle)


def test_trustee_not_owner_cant_create_revoc_reg_entry(looper,
                                                       txnPoolNodeSet,
                                                       sdk_wallet_trustee,
                                                       sdk_wallet_steward,
                                                       sdk_pool_handle,
                                                       build_revoc_def_by_default,
                                                       claim_def, tconf):
    revoc_def_req_steward = create_revoc_reg_def(looper, txnPoolNodeSet, sdk_pool_handle,
                                                 build_revoc_def_by_default,
                                                 claim_def, sdk_wallet_steward)

    rev_reg_entry = build_revoc_reg_entry_for_given_revoc_reg_def(revoc_def_req_steward)
    rev_reg_entry[VALUE][REVOKED] = [1, 2, 3, 4, 5]
    del rev_reg_entry[VALUE][PREV_ACCUM]
    rev_entry_req_trustee = sdk_sign_request_from_dict(looper, sdk_wallet_trustee, rev_reg_entry)
    with pytest.raises(RequestRejectedException):
        sdk_send_and_check([json.dumps(rev_entry_req_trustee)], looper, txnPoolNodeSet, sdk_pool_handle)


def test_allowed_roles_can_edit_revoc_reg_entry(looper,
                                                txnPoolNodeSet,
                                                sdk_wallet_endorser,
                                                sdk_wallet_steward,
                                                sdk_wallet_trustee,
                                                sdk_pool_handle,
                                                build_revoc_def_by_default,
                                                build_revoc_def_by_endorser,
                                                build_revoc_def_by_steward,
                                                claim_def, tconf):
    # endorser
    revoc_entry_req_endorser = create_revoc_reg_entry(looper, txnPoolNodeSet, sdk_pool_handle,
                                                          build_revoc_def_by_endorser,
                                                          claim_def, sdk_wallet_endorser)

    revoc_entry_req_endorser['operation'][VALUE][REVOKED] = [6, 7, 8]
    revoc_entry_req_endorser['operation'][VALUE][PREV_ACCUM] = revoc_entry_req_endorser['operation'][VALUE][
        ACCUM]
    revoc_entry_req_endorser['operation'][VALUE][ACCUM] = randomString(10)
    revoc_entry_req_endorser = sdk_sign_request_from_dict(looper, sdk_wallet_endorser,
                                                              revoc_entry_req_endorser['operation'])

    sdk_send_and_check([json.dumps(revoc_entry_req_endorser)], looper, txnPoolNodeSet, sdk_pool_handle)

    # steward
    revoc_entry_req_trustee = create_revoc_reg_entry(looper, txnPoolNodeSet, sdk_pool_handle,
                                                     build_revoc_def_by_steward,
                                                     claim_def, sdk_wallet_steward)

    revoc_entry_req_trustee['operation'][VALUE][REVOKED] = [6, 7, 8]
    revoc_entry_req_trustee['operation'][VALUE][PREV_ACCUM] = revoc_entry_req_trustee['operation'][VALUE][ACCUM]
    revoc_entry_req_trustee['operation'][VALUE][ACCUM] = randomString(10)
    revoc_entry_req_trustee = sdk_sign_request_from_dict(looper, sdk_wallet_steward,
                                                         revoc_entry_req_trustee['operation'])

    sdk_send_and_check([json.dumps(revoc_entry_req_trustee)], looper, txnPoolNodeSet, sdk_pool_handle)

    # steward
    revoc_entry_req_steward = create_revoc_reg_entry(looper, txnPoolNodeSet, sdk_pool_handle,
                                                     build_revoc_def_by_default,
                                                     claim_def, sdk_wallet_trustee)

    revoc_entry_req_steward['operation'][VALUE][REVOKED] = [6, 7, 8]
    revoc_entry_req_steward['operation'][VALUE][PREV_ACCUM] = revoc_entry_req_steward['operation'][VALUE][ACCUM]

    revoc_entry_req_steward['operation'][VALUE][ACCUM] = randomString(10)
    revoc_entry_req_steward = sdk_sign_request_from_dict(looper, sdk_wallet_trustee,
                                                         revoc_entry_req_steward['operation'])

    sdk_send_and_check([json.dumps(revoc_entry_req_steward)], looper, txnPoolNodeSet, sdk_pool_handle)


def test_not_owner_trustee_cant_edit_revoc_reg_entry(looper,
                                                     txnPoolNodeSet,
                                                     sdk_wallet_trustee,
                                                     sdk_wallet_endorser,
                                                     sdk_pool_handle,
                                                     build_revoc_def_by_demand,
                                                     claim_def, tconf):
    revoc_entry_req_steward = create_revoc_reg_entry(looper, txnPoolNodeSet, sdk_pool_handle,
                                                     build_revoc_def_by_demand,
                                                     claim_def, sdk_wallet_endorser)

    revoc_entry_req_steward['operation'][VALUE][REVOKED] = [6, 7, 8]
    revoc_entry_req_steward['operation'][VALUE][PREV_ACCUM] = revoc_entry_req_steward['operation'][VALUE][ACCUM]

    revoc_entry_req_steward['operation'][VALUE][ACCUM] = randomString(10)
    revoc_entry_req_steward = sdk_sign_request_from_dict(looper, sdk_wallet_trustee,
                                                         revoc_entry_req_steward['operation'])
    with pytest.raises(RequestRejectedException):
        sdk_send_and_check([json.dumps(revoc_entry_req_steward)], looper, txnPoolNodeSet, sdk_pool_handle)


def test_not_owner_steward_cant_edit_revoc_reg_entry(looper,
                                                     txnPoolNodeSet,
                                                     sdk_wallet_steward,
                                                     sdk_wallet_endorser,
                                                     sdk_pool_handle,
                                                     build_revoc_def_by_steward,
                                                     claim_def, tconf):
    revoc_entry_req_steward = create_revoc_reg_entry(looper, txnPoolNodeSet, sdk_pool_handle,
                                                     build_revoc_def_by_steward,
                                                     claim_def, sdk_wallet_endorser)

    revoc_entry_req_steward['operation'][VALUE][REVOKED] = [6, 7, 8]
    revoc_entry_req_steward['operation'][VALUE][PREV_ACCUM] = revoc_entry_req_steward['operation'][VALUE][ACCUM]

    revoc_entry_req_steward['operation'][VALUE][ACCUM] = randomString(10)
    revoc_entry_req_steward = sdk_sign_request_from_dict(looper, sdk_wallet_steward,
                                                         revoc_entry_req_steward['operation'])
    with pytest.raises(RequestRejectedException):
        sdk_send_and_check([json.dumps(revoc_entry_req_steward)], looper, txnPoolNodeSet, sdk_pool_handle)


def test_not_owner_endorser_cant_edit_revoc_reg_entry(looper,
                                                          txnPoolNodeSet,
                                                          sdk_wallet_steward,
                                                          sdk_wallet_endorser,
                                                          sdk_pool_handle,
                                                          build_revoc_def_by_default,
                                                          claim_def, tconf):
    revoc_build = build_revoc_def_random(looper, sdk_wallet_steward)
    revoc_entry_req_steward = create_revoc_reg_entry(looper, txnPoolNodeSet, sdk_pool_handle,
                                                     revoc_build, claim_def, sdk_wallet_steward)

    revoc_entry_req_steward['operation'][VALUE][REVOKED] = [6, 7, 8]
    revoc_entry_req_steward['operation'][VALUE][PREV_ACCUM] = revoc_entry_req_steward['operation'][VALUE][ACCUM]

    revoc_entry_req_steward['operation'][VALUE][ACCUM] = randomString(10)
    revoc_entry_req_steward = sdk_sign_request_from_dict(looper, sdk_wallet_endorser,
                                                         revoc_entry_req_steward['operation'])
    with pytest.raises(RequestRejectedException):
        sdk_send_and_check([json.dumps(revoc_entry_req_steward)], looper, txnPoolNodeSet, sdk_pool_handle)



from typing import Optional

from indy_common.authorize.auth_actions import AuthActionAdd, AuthActionEdit
from indy_common.authorize.auth_request_validator import WriteRequestValidator

from indy_common.constants import SCHEMA, SCHEMA_ATTR_NAMES

from indy_common.req_utils import get_write_schema_name, get_write_schema_version, get_txn_schema_name, \
    get_txn_schema_version, get_txn_schema_attr_names
from indy_common.state.state_constants import MARKER_SCHEMA
from plenum.common.constants import DOMAIN_LEDGER_ID

from plenum.common.request import Request
from plenum.common.txn_util import get_request_data, get_from, get_seq_no, get_txn_time
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.handler_interfaces.write_request_handler import WriteRequestHandler
from plenum.server.request_handlers.utils import encode_state_value


class SchemaHandler(WriteRequestHandler):

    def __init__(self, database_manager: DatabaseManager,
                 write_req_validator: WriteRequestValidator):
        super().__init__(database_manager, SCHEMA, DOMAIN_LEDGER_ID)
        self.write_req_validator = write_req_validator

    def static_validation(self, request: Request):
        pass

    def additional_dynamic_validation(self, request: Request, req_pp_time: Optional[int]):
        # we can not add a Schema with already existent NAME and VERSION
        # sine a Schema needs to be identified by seqNo
        self._validate_request_type(request)
        identifier, req_id, operation = get_request_data(request)
        schema_name = get_write_schema_name(request)
        schema_version = get_write_schema_version(request)
        path = SchemaHandler.make_state_path_for_schema(identifier, schema_name, schema_version)
        schema, _, _ = self.get_from_state(path)
        if schema:
            self.write_req_validator.validate(request,
                                              [AuthActionEdit(txn_type=SCHEMA,
                                                              field='*',
                                                              old_value='*',
                                                              new_value='*')])
        else:
            self.write_req_validator.validate(request,
                                              [AuthActionAdd(txn_type=SCHEMA,
                                                             field='*',
                                                             value='*')])

    def gen_txn_id(self, txn):
        self._validate_txn_type(txn)
        path = SchemaHandler.prepare_schema_for_state(txn, path_only=True)
        return path.decode()

    def update_state(self, txn, prev_result, request, is_committed=False) -> None:
        self._validate_txn_type(txn)
        path, value_bytes = SchemaHandler.prepare_schema_for_state(txn)
        self.state.set(path, value_bytes)
        return txn

    @staticmethod
    def prepare_schema_for_state(txn, path_only=False):
        origin = get_from(txn)
        schema_name = get_txn_schema_name(txn)
        schema_version = get_txn_schema_version(txn)
        value = {
            SCHEMA_ATTR_NAMES: get_txn_schema_attr_names(txn)
        }
        path = SchemaHandler.make_state_path_for_schema(origin, schema_name, schema_version)
        if path_only:
            return path
        seq_no = get_seq_no(txn)
        txn_time = get_txn_time(txn)
        value_bytes = encode_state_value(value, seq_no, txn_time)
        return path, value_bytes

    @staticmethod
    def make_state_path_for_schema(authors_did, schema_name, schema_version) -> bytes:
        return "{DID}:{MARKER}:{SCHEMA_NAME}:{SCHEMA_VERSION}" \
            .format(DID=authors_did,
                    MARKER=MARKER_SCHEMA,
                    SCHEMA_NAME=schema_name,
                    SCHEMA_VERSION=schema_version).encode()



import pytest

from indy_common.authorize.auth_actions import AuthActionAdd, AuthActionEdit
from indy_common.constants import JSON_LD_CONTEXT, RICH_SCHEMA, RICH_SCHEMA_ENCODING, \
    RICH_SCHEMA_MAPPING, RICH_SCHEMA_CRED_DEF, RICH_SCHEMA_PRES_DEF


@pytest.mark.parametrize('txn_type',
                         [JSON_LD_CONTEXT, RICH_SCHEMA, RICH_SCHEMA_ENCODING, RICH_SCHEMA_MAPPING,
                          RICH_SCHEMA_CRED_DEF, RICH_SCHEMA_PRES_DEF])
def test_rich_schema_object_adding(write_request_validation, req, is_owner, txn_type):
    authorized = req.identifier in ("trustee_identifier", "steward_identifier", "endorser_identifier")
    assert authorized == write_request_validation(req,
                                                  [AuthActionAdd(txn_type=txn_type,
                                                                 field='some_field',
                                                                 value='some_value',
                                                                 is_owner=is_owner)])


@pytest.mark.parametrize('txn_type',
                         [JSON_LD_CONTEXT, RICH_SCHEMA, RICH_SCHEMA_ENCODING, RICH_SCHEMA_MAPPING,
                          RICH_SCHEMA_CRED_DEF, RICH_SCHEMA_PRES_DEF])
def test_rich_schema_object_editing(write_request_validation, req, is_owner, txn_type):
    assert not write_request_validation(req,
                                        [AuthActionEdit(txn_type=txn_type,
                                                        field='some_field',
                                                        old_value='old_value',
                                                        new_value='new_value',
                                                        is_owner=is_owner)])



from indy_common.authorize.auth_actions import AuthActionEdit, EDIT_PREFIX, AuthActionAdd, ADD_PREFIX
from indy_common.authorize.auth_constraints import ConstraintCreator
from indy_common.constants import AUTH_ACTION, OLD_VALUE, NEW_VALUE, AUTH_TYPE, FIELD, CONSTRAINT
from plenum.common.exceptions import InvalidClientRequest


class StaticAuthRuleHelper:

    @staticmethod
    def get_auth_key(operation):
        action = operation.get(AUTH_ACTION, None)
        old_value = operation.get(OLD_VALUE, None)
        new_value = operation.get(NEW_VALUE, None)
        auth_type = operation.get(AUTH_TYPE, None)
        field = operation.get(FIELD, None)

        return AuthActionEdit(txn_type=auth_type,
                              field=field,
                              old_value=old_value,
                              new_value=new_value).get_action_id() \
            if action == EDIT_PREFIX else \
            AuthActionAdd(txn_type=auth_type,
                          field=field,
                          value=new_value).get_action_id()

    @staticmethod
    def get_auth_constraint(operation):
        return ConstraintCreator.create_constraint(operation.get(CONSTRAINT))

    @staticmethod
    def check_auth_key(operation, identifier, req_id, auth_map):
        action = operation.get(AUTH_ACTION, None)

        if OLD_VALUE not in operation and action == EDIT_PREFIX:
            raise InvalidClientRequest(identifier, req_id,
                                       "Transaction for change authentication "
                                       "rule for {}={} must contain field {}".
                                       format(AUTH_ACTION, EDIT_PREFIX, OLD_VALUE))

        if OLD_VALUE in operation and action == ADD_PREFIX:
            raise InvalidClientRequest(identifier, req_id,
                                       "Transaction for change authentication "
                                       "rule for {}={} must not contain field {}".
                                       format(AUTH_ACTION, ADD_PREFIX, OLD_VALUE))

        auth_key = StaticAuthRuleHelper.get_auth_key(operation)

        if auth_key not in auth_map:
            raise InvalidClientRequest(identifier, req_id,
                                       "Unknown authorization rule: key '{}' is not "
                                       "found in authorization map.".format(auth_key))



import json

import pytest
from indy.ledger import build_attrib_request, sign_request

from indy_common.types import Request
from plenum.common.request import Request as PRequest

from plenum.common.txn_util import reqToTxn, get_digest, get_payload_digest
from plenum.test.helper import sdk_set_protocol_version


@pytest.fixture(scope="module")
def req_json(looper, sdk_wallet_steward):
    sdk_set_protocol_version(looper)
    wallet_handle, identifier = sdk_wallet_steward
    raw = json.dumps({'answer': 42})
    return looper.loop.run_until_complete(
        build_attrib_request(identifier, identifier, raw=raw, xhash=None, enc=None))


@pytest.fixture(scope="module")
def req(req_json, looper, sdk_wallet_steward):
    wallet_handle, identifier = sdk_wallet_steward
    req_signed = looper.loop.run_until_complete(sign_request(wallet_handle, identifier, req_json))
    return Request(**json.loads(req_signed))


def test_attrib_txn_payload_digest_req_json(req_json, req):
    txn = reqToTxn(req_json)
    assert get_payload_digest(txn) == req.payload_digest


def test_attrib_txn_digest_req_dict(req):
    txn = reqToTxn(req.as_dict)
    assert get_digest(txn) == req.digest


def test_attrib_txn_digest_req_instance(req):
    txn = reqToTxn(req)
    assert get_digest(txn) == req.digest


def test_attrib_txn_different_payload(req_json):
    req_json = json.loads(req_json)
    n_req = Request(**req_json)
    p_req = PRequest(**req_json)
    assert n_req.digest != p_req.payload_digest
    assert n_req.payload_digest != p_req.payload_digest



import pytest

from common.version import InvalidVersionError
from plenum.common.exceptions import BlowUp

from indy_common.constants import APP_NAME
from indy_common.version import src_version_cls

from indy_node.test.upgrade.helper import (
    NodeControlToolPatched,
    composeUpgradeMessage
)


@pytest.fixture(scope='module')
def tconf(tconf, tdir):
    tconf.LOG_DIR = tdir
    yield tconf


def test_node_control_tool_processes_invalid_json(
    monkeypatch, tdir, tconf
):
    def patch(tool):
        monkeypatch.setattr(tool, '_listen', lambda *x, **y: None)
        monkeypatch.setattr(tool, '_upgrade', lambda *x, **y: None)
        monkeypatch.setattr(tool, '_restart', lambda *x, **y: None)

    tool = NodeControlToolPatched(patch, backup_dir=tdir, backup_target=tdir, config=tconf)

    with pytest.raises(BlowUp, match='JSON decoding failed'):
        tool._process_data('{12345}'.encode('utf-8'))


@pytest.mark.parametrize(
    'pkg_name,version',
    [
        ('some-pkg', '1.2.3.4.5'),
        (APP_NAME, '1.2.3.4.5'),
    ]
)
def test_node_control_tool_processes_invalid_version(
    monkeypatch, tdir, tconf, pkg_name, version
):
    def patch(tool):
        monkeypatch.setattr(tool, '_listen', lambda *x, **y: None)
        monkeypatch.setattr(tool, '_upgrade', lambda *x, **y: None)
        monkeypatch.setattr(tool, '_restart', lambda *x, **y: None)

    tool = NodeControlToolPatched(patch, backup_dir=tdir, backup_target=tdir, config=tconf)

    with pytest.raises(InvalidVersionError):
        src_version_cls(pkg_name)(version)

    with pytest.raises(BlowUp, match='invalid version 1.2.3.4.5'):
        tool._process_data(composeUpgradeMessage(version, pkg_name))



from indy_node.server.upgrade_log import UpgradeLog
from indy_node.test import waits
from indy_node.test.upgrade.helper import checkUpgradeScheduled, \
    sdk_ensure_upgrade_sent, count_action_log_package
from plenum.common.constants import VERSION
from plenum.test.delayers import req_delay
from plenum.test.test_node import getNonPrimaryReplicas
from stp_core.loop.eventually import eventually


def test_forced_upgrade_handled_once_if_ordered_and_then_request_received(
        looper, nodeSet, sdk_pool_handle, sdk_wallet_trustee,
        validUpgradeExpForceTrue):
    """
    Verifies that POOL_UPGRADE force=true request is handled one time in case
    the node commits the transaction to the ledger and only after that receives
    the request directly from the client
    """
    slow_node = getNonPrimaryReplicas(nodeSet, instId=0)[-1].node
    slow_node.clientIbStasher.delay(req_delay())

    sdk_ensure_upgrade_sent(looper, sdk_pool_handle, sdk_wallet_trustee,
                            validUpgradeExpForceTrue)

    looper.run(eventually(checkUpgradeScheduled,
                          [slow_node],
                          validUpgradeExpForceTrue[VERSION],
                          retryWait=1,
                          timeout=waits.expectedUpgradeScheduled()))

    slow_node.clientIbStasher.reset_delays_and_process_delayeds()
    looper.runFor(waits.expectedUpgradeScheduled())

    checkUpgradeScheduled([slow_node], validUpgradeExpForceTrue[VERSION])
    assert count_action_log_package(list(slow_node.upgrader._actionLog), validUpgradeExpForceTrue['package']) == 1
    assert slow_node.upgrader._actionLog.last_event.ev_type == UpgradeLog.Events.scheduled



import pytest

from indy_common.constants import ENDORSER, NETWORK_MONITOR
from indy_common.test.constants import IDENTIFIERS, OTHER_ROLE
from indy_common.types import Request
from indy_node.server.request_handlers.config_req_handlers.txn_author_agreement_disable_handler import \
    TxnAuthorAgreementDisableHandler
from indy_node.test.conftest import write_auth_req_validator, idr_cache
from plenum.common.constants import TXN_TYPE, TXN_AUTHOR_AGREEMENT_DISABLE, TRUSTEE, STEWARD
from plenum.common.exceptions import UnauthorizedClientRequest


@pytest.fixture()
def disable_taa_handler(db_manager, write_auth_req_validator):
    return TxnAuthorAgreementDisableHandler(database_manager=db_manager,
                                            write_req_validator=write_auth_req_validator)


@pytest.fixture(params=[TRUSTEE, ENDORSER, STEWARD, NETWORK_MONITOR, None, OTHER_ROLE])
def disable_taa_request(request):
    operation = {TXN_TYPE: TXN_AUTHOR_AGREEMENT_DISABLE}
    return (request.param, Request(identifier=IDENTIFIERS[request.param][0],
                                   signature="sign",
                                   operation=operation))

def test_taa_disable_only_for_trustee(disable_taa_handler,
                                      disable_taa_request):
    who, req = disable_taa_request
    if who != TRUSTEE:
        with pytest.raises(UnauthorizedClientRequest, match="Not enough TRUSTEE signatures"):
            disable_taa_handler.authorize(req)
    else:
        disable_taa_handler.authorize(req)


import pytest

from indy_node.test.suspension.helper import sdk_suspend_role
from indy_node.test.suspension.test_suspension import another_trustee

from plenum.common.exceptions import RequestRejectedException, RequestNackedException
from plenum.test.pool_transactions.helper import sdk_add_new_nym


def testTrusteeSuspendingEndorser(looper, sdk_pool_handle, sdk_wallet_trustee,
                                     sdk_wallet_endorser):
    _, did = sdk_wallet_endorser
    sdk_suspend_role(looper, sdk_pool_handle, sdk_wallet_trustee, did)
    with pytest.raises(RequestRejectedException) as e:
        sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_endorser)
    e.match('Rule for this action is')


def testTrusteeSuspendingTrustee(looper, sdk_pool_handle, sdk_wallet_trustee,
                                 another_trustee):
    _, did = another_trustee
    sdk_suspend_role(looper, sdk_pool_handle, sdk_wallet_trustee, did)
    with pytest.raises(RequestRejectedException) as e:
        sdk_add_new_nym(looper, sdk_pool_handle, another_trustee)
    e.match('Rule for this action is')


def testTrusteeSuspendingSteward(looper, sdk_pool_handle, sdk_wallet_trustee,
                                 sdk_wallet_steward):
    _, did = sdk_wallet_steward
    sdk_suspend_role(looper, sdk_pool_handle, sdk_wallet_trustee, did)
    with pytest.raises(RequestRejectedException) as e:
        sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_steward)
    e.match('Rule for this action is')


def testEndorserSuspendingHimselfByVerkeyFlush(looper, sdk_pool_handle,
                                                  sdk_wallet_endorser):
    # The endorser has already lost its role due to previous tests,
    # but it is ok for this test where the endorser flushes its verkey
    # and then he is unable to send NYM due to empty verkey.
    _, did = sdk_wallet_endorser
    sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_endorser, dest=did, verkey='')
    with pytest.raises(RequestNackedException) as e:
        sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_endorser)
    e.match('insufficient number of valid signatures')



from indy_common.authorize.auth_actions import AuthActionEdit
from indy_common.authorize.auth_constraints import CONSTRAINT_ID, ConstraintsEnum, OFF_LEDGER_SIGNATURE, \
    AUTH_CONSTRAINTS
from indy_common.authorize.auth_request_validator import WriteRequestValidator
from indy_common.constants import AUTH_RULE
from indy_node.server.request_handlers.config_req_handlers.auth_rule.abstract_auth_rule_handler import \
    AbstractAuthRuleHandler
from indy_node.server.request_handlers.config_req_handlers.auth_rule.auth_rule_handler import AuthRuleHandler
from indy_node.server.request_handlers.config_req_handlers.auth_rule.static_auth_rule_helper import StaticAuthRuleHelper
from plenum.common.request import Request
from plenum.common.txn_util import get_payload_data
from plenum.server.database_manager import DatabaseManager


class AuthRuleHandler191(AuthRuleHandler):

    def _update_auth_constraint(self, auth_key: str, constraint):
        constraint_dict = constraint.as_dict if constraint else {}
        self._set_off_ledger_to_constraint(constraint_dict)
        self.state.set(AbstractAuthRuleHandler.make_state_path_for_auth_rule(auth_key),
                       self.constraint_serializer.serializer.serialize(constraint_dict))

    def _set_off_ledger_to_constraint(self, constraint_dict):
        if constraint_dict[CONSTRAINT_ID] == ConstraintsEnum.ROLE_CONSTRAINT_ID and \
                not constraint_dict.get(OFF_LEDGER_SIGNATURE, False):
            constraint_dict[OFF_LEDGER_SIGNATURE] = False
        for constraint in constraint_dict.get(AUTH_CONSTRAINTS, []):
            self._set_off_ledger_to_constraint(constraint)



import pytest
from collections import OrderedDict
from plenum.common.messages.fields import ConstantField, ChooseField, MapField, Sha256HexField, \
    NonNegativeNumberField, LimitedLengthStringField, BooleanField
from plenum.common.messages.fields import VersionField
from indy_common.types import ClientPoolUpgradeOperation


EXPECTED_ORDERED_FIELDS = OrderedDict([
    ("type", ConstantField),
    ('action', ChooseField),
    ("version", VersionField),
    ('schedule', MapField),
    ('sha256', Sha256HexField),
    ('timeout', NonNegativeNumberField),
    ('justification', LimitedLengthStringField),
    ("name", LimitedLengthStringField),
    ("force", BooleanField),
    ("reinstall", BooleanField),
    ("package", LimitedLengthStringField),
])


def test_has_expected_fields():
    actual_field_names = OrderedDict(ClientPoolUpgradeOperation.schema).keys()
    assert actual_field_names == EXPECTED_ORDERED_FIELDS.keys()


def test_has_expected_validators():
    schema = dict(ClientPoolUpgradeOperation.schema)
    for field, validator in EXPECTED_ORDERED_FIELDS.items():
        assert isinstance(schema[field], validator)



from copy import deepcopy
import pytest

from plenum.common.exceptions import RequestNackedException, RequestRejectedException
from plenum.common.constants import VERSION
from plenum.common.util import randomString
from indy_node.test.upgrade.helper import loweredVersion, sdk_ensure_upgrade_sent
from indy_common.constants import JUSTIFICATION, JUSTIFICATION_MAX_SIZE

from indy_node.test.upgrade.conftest import validUpgrade, nodeIds, pckg


def testPoolUpgradeFailsIfVersionIsLowerThanCurrent(
        looper, sdk_pool_handle, validUpgrade, sdk_wallet_trustee):
    upgrade = deepcopy(validUpgrade)
    upgrade[VERSION] = loweredVersion()

    with pytest.raises(RequestRejectedException) as e:
        sdk_ensure_upgrade_sent(looper, sdk_pool_handle, sdk_wallet_trustee, upgrade)
    e.match('Version {} is not upgradable'.format(upgrade[VERSION]))


def testPoolUpgradeHasInvalidSyntaxIfJustificationIsEmpty(
        looper, sdk_pool_handle, validUpgrade, sdk_wallet_trustee):
    upgrade = deepcopy(validUpgrade)
    upgrade[JUSTIFICATION] = ''

    with pytest.raises(RequestNackedException) as e:
        sdk_ensure_upgrade_sent(looper, sdk_pool_handle, sdk_wallet_trustee, upgrade)
    e.match('empty string')


def testPoolUpgradeHasInvalidSyntaxIfJustificationIsVeryLong(
        looper, sdk_pool_handle, validUpgrade, sdk_wallet_trustee):
    upgrade = deepcopy(validUpgrade)
    upgrade[JUSTIFICATION] = randomString(JUSTIFICATION_MAX_SIZE + 1)

    with pytest.raises(RequestNackedException) as e:
        sdk_ensure_upgrade_sent(looper, sdk_pool_handle, sdk_wallet_trustee, upgrade)
    e.match('is longer than {} symbols'.format(JUSTIFICATION_MAX_SIZE))



import time

from indy_common.constants import TIMESTAMP
from indy_common.types import SafeRequest
from indy_node.test.anon_creds.helper import build_get_revoc_reg_entry
from plenum.common.constants import TXN_TIME, STATE_PROOF, DATA, DOMAIN_LEDGER_ID
from plenum.common.types import f, OPERATION


def test_validate_get_revoc_reg_entry(looper, sdk_wallet_steward):
    req = build_get_revoc_reg_entry(looper, sdk_wallet_steward)
    SafeRequest(**req)


def test_get_revoc_reg_entry_without_any_rev_entry(send_revoc_reg_def_by_default,
                                                   looper, sdk_wallet_steward,
                                                   txnPoolNodeSet):
    req = build_get_revoc_reg_entry(looper, sdk_wallet_steward)
    req[OPERATION][TIMESTAMP] = int(time.time()) + 2
    result = txnPoolNodeSet[0].read_manager.get_result((SafeRequest(**req)))
    assert result[DATA] is None
    assert result[f.SEQ_NO.nm] is None
    assert result[TXN_TIME] is None
    assert STATE_PROOF in result
    assert result[STATE_PROOF] is not None



from binascii import hexlify
from hashlib import sha256
import json
from typing import Mapping, Optional

import base58
from common.serializers.serialization import domain_state_serializer
from ledger.util import F
from plenum.common.constants import ROLE, TARGET_NYM, TXN_TIME, VERKEY
from plenum.common.exceptions import InvalidClientRequest
from plenum.common.request import Request
from plenum.common.txn_util import (
    get_from,
    get_payload_data,
    get_request_data,
    get_seq_no,
    get_txn_time,
)
from plenum.common.types import f
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.nym_handler import NymHandler as PNymHandler
from plenum.server.request_handlers.utils import get_nym_details, nym_to_state_key

from indy_common.auth import Authoriser
from indy_common.authorize.auth_actions import AuthActionAdd, AuthActionEdit
from indy_common.authorize.auth_request_validator import WriteRequestValidator
from indy_common.constants import (
    DIDDOC_CONTENT,
    NYM,
    NYM_VERSION,
    NYM_VERSION_CONVENTION,
    NYM_VERSION_NULL,
    NYM_VERSION_SELF_CERT,
)
from indy_common.exceptions import InvalidDIDDocException


class NymHandler(PNymHandler):
    state_serializer = domain_state_serializer

    def __init__(
        self,
        config,
        database_manager: DatabaseManager,
        write_req_validator: WriteRequestValidator,
    ):
        super().__init__(config, database_manager)
        self.write_req_validator = write_req_validator

    def static_validation(self, request: Request):
        self._validate_request_type(request)
        identifier, req_id, operation = get_request_data(request)
        assert isinstance(operation, Mapping)
        role = operation.get(ROLE)
        nym = operation.get(TARGET_NYM)
        if isinstance(nym, str):
            nym = nym.strip()
        if not nym:
            raise InvalidClientRequest(
                identifier, req_id, "{} needs to be present".format(TARGET_NYM)
            )
        if not Authoriser.isValidRole(role):
            raise InvalidClientRequest(
                identifier, req_id, "{} not a valid role".format(role)
            )

        diddoc_content = operation.get(DIDDOC_CONTENT, None)
        if diddoc_content:
            diddoc_content = json.loads(diddoc_content)
            try:
                self._validate_diddoc_content(diddoc_content)
            except InvalidDIDDocException as error:
                raise InvalidClientRequest(
                    identifier,
                    req_id,
                    error.reason,
                ) from error

        version = operation.get(NYM_VERSION)
        if version is not None and version not in (
            NYM_VERSION_NULL,
            NYM_VERSION_CONVENTION,
            NYM_VERSION_SELF_CERT,
        ):
            raise InvalidClientRequest(
                request.identifier,
                request.reqId,
                "Version must be one of {{{}}}, received {}".format(
                    ", ".join(
                        str(val)
                        for val in [
                            NYM_VERSION_NULL,
                            NYM_VERSION_CONVENTION,
                            NYM_VERSION_SELF_CERT,
                        ]
                    ),
                    version
                ),
            )

    def additional_dynamic_validation(
        self, request: Request, req_pp_time: Optional[int]
    ):
        self._validate_request_type(request)
        operation = request.operation
        assert isinstance(operation, Mapping)
        nym_data = self.database_manager.idr_cache.getNym(
            operation[TARGET_NYM], isCommitted=False
        )

        if nym_data:
            self._validate_existing_nym(request, operation, nym_data)
        else:
            self._validate_new_nym(request, operation)

    def gen_txn_id(self, txn):
        self._validate_txn_type(txn)
        nym = get_payload_data(txn).get(TARGET_NYM)
        binary_digest = self.make_state_path_for_nym(nym)
        return hexlify(binary_digest).decode()

    def update_state(self, txn, prev_result, request, is_committed=False):
        self._validate_txn_type(txn)
        nym = get_payload_data(txn).get(TARGET_NYM)
        existing_data = get_nym_details(self.state, nym, is_committed=is_committed)
        txn_data = get_payload_data(txn)
        new_data = {}
        if not existing_data:
            new_data[f.IDENTIFIER.nm] = get_from(txn)
            new_data[ROLE] = None
            new_data[VERKEY] = None

        if ROLE in txn_data:
            new_data[ROLE] = txn_data[ROLE]
        if VERKEY in txn_data:
            new_data[VERKEY] = txn_data[VERKEY]
        if DIDDOC_CONTENT in txn_data:
            new_data[DIDDOC_CONTENT] = txn_data[DIDDOC_CONTENT]
        if NYM_VERSION in txn_data:
            new_data[NYM_VERSION] = txn_data[NYM_VERSION]
        new_data[F.seqNo.name] = get_seq_no(txn)
        new_data[TXN_TIME] = get_txn_time(txn)
        existing_data.update(new_data)
        val = self.state_serializer.serialize(existing_data)
        key = nym_to_state_key(nym)
        self.state.set(key, val)
        return existing_data

    def _validate_new_nym(self, request, operation):
        identifier, req_id, _ = get_request_data(request)
        role = operation.get(ROLE)

        nym_data = self.database_manager.idr_cache.getNym(
            request.identifier, isCommitted=False
        )
        if not nym_data:
            # Non-ledger nym case. These two checks duplicated and mainly executed in client_authn,
            # but it has point to repeat them here, for clear understanding of validation non-ledger request cases.
            if request.identifier != request.operation.get(TARGET_NYM):
                raise InvalidClientRequest(
                    identifier,
                    req_id,
                    "DID which is not stored on ledger can "
                    "send nym txn only if appropriate auth_rules set "
                    "and sender did equal to destination nym",
                )
            if not request.operation.get(VERKEY):
                raise InvalidClientRequest(
                    identifier,
                    req_id,
                    "Non-ledger nym txn must contain verkey for new did",
                )

        version = request.operation.get(NYM_VERSION)
        if version == NYM_VERSION_CONVENTION and not self._legacy_convention_validation(
            request.operation.get(TARGET_NYM), request.operation.get(VERKEY)
        ):
            raise InvalidClientRequest(
                identifier,
                req_id,
                "Identifier with version 1 must be first 16 bytes of verkey",
            )
        elif version == NYM_VERSION_SELF_CERT and not self._is_self_certifying(
            request.operation.get(TARGET_NYM), request.operation.get(VERKEY)
        ):
            raise InvalidClientRequest(
                identifier,
                req_id,
                "Identifier with version 2 must be first 16 bytes of SHA256 of verkey",
            )

        self.write_req_validator.validate(
            request, [AuthActionAdd(txn_type=NYM, field=ROLE, value=role)]
        )

    def _validate_existing_nym(self, request, operation, nym_data):
        origin = request.identifier
        owner = self.database_manager.idr_cache.getOwnerFor(
            operation[TARGET_NYM], isCommitted=False
        )
        is_owner = origin == owner

        updateKeys = [ROLE, VERKEY]
        updateKeysInOperationOrOwner = is_owner

        if NYM_VERSION in operation:
            raise InvalidClientRequest(
                request.identifier,
                request.reqId,
                "Cannot set version on existing nym"
            )

        for key in updateKeys:
            if key in operation:
                updateKeysInOperationOrOwner = True
                newVal = operation[key]
                oldVal = nym_data.get(key)
                self.write_req_validator.validate(
                    request,
                    [
                        AuthActionEdit(
                            txn_type=NYM,
                            field=key,
                            old_value=oldVal,
                            new_value=newVal,
                            is_owner=is_owner,
                        )
                    ],
                )
        if not updateKeysInOperationOrOwner:
            raise InvalidClientRequest(request.identifier, request.reqId)

    def _decode_state_value(self, encoded):
        if encoded:
            return domain_state_serializer.deserialize(encoded)
        return None, None, None

    def _is_self_certifying(self, identifier: str, verkey: str):
        """Validate that the identifier is self-certifying.

        DID must be the base58 encoded first 16 bytes of the 32 bytes verkey
        See https://hyperledger.github.io/indy-did-method/#creation
        """
        return identifier == base58.b58encode(
            sha256(base58.b58decode(verkey)).digest()[:16]
        ).decode("utf-8")

    def _legacy_convention_validation(self, identifier: str, verkey: str):
        """Validate old Indy SDK convention for DID generation.

        The did is derived from the first 16 bytes of the verkey.
        """
        return identifier == base58.b58encode(
            base58.b58decode(verkey)[:16]
        ).decode("utf-8")

    def _validate_diddoc_content(self, diddoc):
        """Validate DID Doc according to assembly rules.

        See https://hyperledger.github.io/indy-did-method/#creation
        """

        # Must not have an id property
        if diddoc.get("id", None):
            raise InvalidDIDDocException(
                "diddocContent must not have `id` at root"
            )

        # No element in diddoc is allowed to have same id as verkey in base
        # diddoc
        for el in diddoc.values():
            if isinstance(el, list):
                for item in el:
                    if self._has_same_id_fragment(item, "verkey"):
                        raise InvalidDIDDocException(
                            "diddocContent must not have object with `id` "
                            "containing fragment `verkey`"
                        )

    def _has_same_id_fragment(self, item, fragment):
        return (
            isinstance(item, dict)
            and "id" in item
            and item["id"].partition("#")[2] == fragment
        )



import pytest

from indy_node.server.request_handlers.action_req_handlers.validator_info_handler import ValidatorInfoHandler
from indy_node.test.validator_info.helper import sdk_get_validator_info

from indy_common.authorize.auth_actions import ADD_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraint, IDENTITY_OWNER
from indy_common.constants import VALIDATOR_INFO
from plenum.common.exceptions import RequestRejectedException
from plenum.test.pool_transactions.helper import sdk_add_new_nym
from plenum.test.testing_utils import FakeSomething

from indy_node.test.helper import build_auth_rule_request_json
from indy_node.test.auth_rule.auth_framework.basic import AuthTest


class ValidatorInfoTest(AuthTest):
    def __init__(self, env, action_id):
        super().__init__(env, action_id)

    def prepare(self):
        self.default_auth_rule = self.get_default_auth_rule()
        self.changed_auth_rule = self.get_changed_auth_rule()
        for n in self.env.txnPoolNodeSet:
            info_tool = FakeSomething(
                info={},
                memory_profiler={},
                _generate_software_info=lambda *args, **kwargs: {},
                extractions={},
                node_disk_size={}
            )
            for h in n.action_manager.request_handlers.values():
                if isinstance(h, ValidatorInfoHandler):
                    h.info_tool = info_tool

    def run(self):
        # Step 1. Check default auth rule
        sdk_get_validator_info(self.looper, self.trustee_wallet, self.sdk_pool_handle)
        with pytest.raises(RequestRejectedException):
            sdk_get_validator_info(self.looper, self.new_default_wallet, self.sdk_pool_handle)

        # Step 2. Change auth rule
        self.send_and_check(self.changed_auth_rule, self.trustee_wallet)

        # Step 3. Check, that we cannot do txn the old way
        with pytest.raises(RequestRejectedException):
            sdk_get_validator_info(self.looper, self.trustee_wallet, self.sdk_pool_handle)

        # Step 4. Check, that new auth rule is used
        sdk_get_validator_info(self.looper, self.new_default_wallet, self.sdk_pool_handle)

        # Step 5. Return default auth rule
        self.send_and_check(self.default_auth_rule, self.trustee_wallet)

        # Step 6. Check, that default auth rule works
        sdk_get_validator_info(self.looper, self.trustee_wallet, self.sdk_pool_handle)
        with pytest.raises(RequestRejectedException):
            sdk_get_validator_info(self.looper, self.new_default_wallet, self.sdk_pool_handle)

    def result(self):
        pass

    def get_changed_auth_rule(self):
        self.new_default_wallet = sdk_add_new_nym(self.looper, self.sdk_pool_handle, self.trustee_wallet, role=IDENTITY_OWNER)
        constraint = AuthConstraint(role=IDENTITY_OWNER,
                                    sig_count=1,
                                    need_to_be_owner=False)
        return build_auth_rule_request_json(
            self.looper, self.trustee_wallet[1],
            auth_action=ADD_PREFIX,
            auth_type=VALIDATOR_INFO,
            field='*',
            new_value='*',
            constraint=constraint.as_dict
        )



from indy_common.authorize.auth_actions import AuthActionEdit
from indy_common.constants import POOL_CONFIG, ACTION


def test_pool_config_change(write_request_validation, is_owner, req):
    authorized = req.identifier == "trustee_identifier"
    assert authorized == write_request_validation(req,
                                                  [AuthActionEdit(txn_type=POOL_CONFIG,
                                                                  field=ACTION,
                                                                  old_value='old_value',
                                                                  new_value='new_value',
                                                                  is_owner=is_owner)])



from copy import deepcopy

import dateutil
from datetime import timedelta

from indy_common.constants import SHA256, SCHEDULE
from indy_node.test import waits
from indy_node.test.upgrade.helper import bumpVersion, checkUpgradeScheduled, \
    sdk_ensure_upgrade_sent
from plenum.common.constants import VERSION, NAME
from plenum.common.util import randomString
from stp_core.loop.eventually import eventually


def test_node_reschedules_upgrade_for_proper_datetime(
        looper, tconf, nodeSet, validUpgrade,
        sdk_pool_handle, sdk_wallet_trustee):
    upgr1 = deepcopy(validUpgrade)
    # Upgrade 1 of each node will be scheduled for tomorrow
    for nodeId in upgr1[SCHEDULE]:
        nodeUpgradeDateTime = dateutil.parser.parse(upgr1[SCHEDULE][nodeId])
        nodeUpgradeDateTime += timedelta(days=1)
        upgr1[SCHEDULE][nodeId] = nodeUpgradeDateTime.isoformat()

    upgr2 = deepcopy(upgr1)
    upgr2[VERSION] = bumpVersion(upgr1[VERSION])
    upgr2[NAME] += randomString(32)
    upgr2[SHA256] = 'ef9c3984e7a31994d4f692139116120bd0dd1ff7e270b6a2d773f8f2f9214d4c'
    # Upgrade 2 of each node will be scheduled for its own day
    # (since today with one day step)
    deltaDays = -1
    for nodeId in upgr2[SCHEDULE]:
        nodeUpgradeDateTime = dateutil.parser.parse(upgr2[SCHEDULE][nodeId])
        nodeUpgradeDateTime += timedelta(days=deltaDays)
        upgr2[SCHEDULE][nodeId] = nodeUpgradeDateTime.isoformat()
        deltaDays += 1

    # Upgrade 1 is scheduled
    sdk_ensure_upgrade_sent(looper, sdk_pool_handle,
                            sdk_wallet_trustee, upgr1)
    looper.run(
        eventually(
            checkUpgradeScheduled,
            nodeSet,
            upgr1[VERSION],
            retryWait=1,
            timeout=waits.expectedUpgradeScheduled()))

    # Upgrade 2 cancels Upgrade 1 and is scheduled itself
    # according its own schedule
    sdk_ensure_upgrade_sent(looper, sdk_pool_handle,
                            sdk_wallet_trustee, upgr2)
    looper.run(
        eventually(
            checkUpgradeScheduled,
            nodeSet,
            upgr2[VERSION],
            upgr2[SCHEDULE],
            retryWait=1,
            timeout=waits.expectedUpgradeScheduled()))



import pytest
from plenum.common.constants import DATA
from plenum.common.exceptions import RequestRejectedException
from plenum.test.freeze_ledgers.helper import sdk_get_frozen_ledgers, sdk_send_freeze_ledgers
from plenum.test.helper import freshness

FRESHNESS_TIMEOUT = 5


@pytest.fixture(scope="module")
def tconf(tconf):
    with freshness(tconf, enabled=True, timeout=FRESHNESS_TIMEOUT):
        yield tconf


def test_send_freeze_ledgers(looper, txnPoolNodeSet, sdk_pool_handle, sdk_wallet_trustee_list):
    with pytest.raises(RequestRejectedException, match="Not enough TRUSTEE signatures"):
        sdk_send_freeze_ledgers(
            looper, sdk_pool_handle,
            sdk_wallet_trustee_list[:-1],
            []
        )

    # check that the config state doesn't contain frozen ledgers records
    result = sdk_get_frozen_ledgers(looper, sdk_pool_handle,
                                    sdk_wallet_trustee_list[0])[1]["result"][DATA]
    assert result is None

    # add to the config state a frozen ledgers record with an empty list
    sdk_send_freeze_ledgers(
        looper, sdk_pool_handle,
        sdk_wallet_trustee_list,
        []
    )

    # check that the config state contains a frozen ledgers record with an empty list
    result = sdk_get_frozen_ledgers(looper, sdk_pool_handle,
                                    sdk_wallet_trustee_list[0])[1]["result"][DATA]
    assert len(result) == 0



from copy import deepcopy

import dateutil
from datetime import timedelta

from indy_common.constants import SCHEDULE
from indy_node.test import waits
from indy_node.test.upgrade.helper import checkUpgradeScheduled, \
    sdk_ensure_upgrade_sent
from plenum.common.constants import VERSION
from stp_core.loop.eventually import eventually


def test_node_schedules_upgrade_for_proper_datetime(
        looper, tconf, nodeSet, validUpgrade, sdk_pool_handle, sdk_wallet_trustee):
    upgr1 = deepcopy(validUpgrade)
    # Upgrade of each node will be scheduled for its own day
    # (since today with one day step)
    deltaDays = 0
    for nodeId in upgr1[SCHEDULE]:
        nodeUpgradeDateTime = dateutil.parser.parse(upgr1[SCHEDULE][nodeId])
        nodeUpgradeDateTime += timedelta(days=deltaDays)
        upgr1[SCHEDULE][nodeId] = nodeUpgradeDateTime.isoformat()
        deltaDays += 1

    # Upgrade is scheduled for the proper datetime for each node
    sdk_ensure_upgrade_sent(looper, sdk_pool_handle,
                            sdk_wallet_trustee, upgr1)
    looper.run(
        eventually(
            checkUpgradeScheduled,
            nodeSet,
            upgr1[VERSION],
            upgr1[SCHEDULE],
            retryWait=1,
            timeout=waits.expectedUpgradeScheduled()))



import json

import pytest
from indy.ledger import parse_get_schema_response, build_get_schema_request

from indy_common.config import SCHEMA_ATTRIBUTES_LIMIT
from indy_node.test.api.helper import sdk_write_schema
from plenum.common.util import randomString
from plenum.config import NAME_FIELD_LIMIT
from plenum.test.helper import sdk_sign_and_submit_req, sdk_get_reply


@pytest.fixture(scope="module")
def schema_json(looper, sdk_pool_handle, sdk_wallet_trustee):
    wallet_handle, identifier = sdk_wallet_trustee
    schema_json, _ = sdk_write_schema(looper, sdk_pool_handle, sdk_wallet_trustee)
    schema_id = json.loads(schema_json)['id']

    request = looper.loop.run_until_complete(build_get_schema_request(identifier, schema_id))
    reply = sdk_get_reply(looper, sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request))[1]
    _, schema_json = looper.loop.run_until_complete(parse_get_schema_response(json.dumps(reply)))
    return schema_json


@pytest.fixture(scope="module")
def large_schema_json(looper, sdk_pool_handle, sdk_wallet_trustee):
    wallet_handle, identifier = sdk_wallet_trustee
    attrs = [randomString(size=NAME_FIELD_LIMIT) for _ in range(SCHEMA_ATTRIBUTES_LIMIT)]
    schema_json, _ = sdk_write_schema(looper, sdk_pool_handle, sdk_wallet_trustee,
                                      multi_attribute=attrs, name="name_large", version="2.0")
    schema_id = json.loads(schema_json)['id']

    request = looper.loop.run_until_complete(build_get_schema_request(identifier, schema_id))
    reply = sdk_get_reply(looper, sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, request))[1]
    _, schema_json = looper.loop.run_until_complete(parse_get_schema_response(json.dumps(reply)))
    return schema_json



from indy_common.authorize.auth_actions import AuthActionAdd, AuthActionEdit

from indy_common.constants import CLAIM_DEF


def test_claim_def_adding(write_request_validation, req, is_owner):
    authorized = req.identifier in ("trustee_identifier", "steward_identifier", "endorser_identifier")
    assert authorized == write_request_validation(req,
                                                  [AuthActionAdd(txn_type=CLAIM_DEF,
                                                                 field='some_field',
                                                                 value='some_value',
                                                                 is_owner=is_owner)])


def test_claim_def_editing(write_request_validation, req, is_owner):
    authorized = is_owner
    assert authorized == write_request_validation(req,
                                                  [AuthActionEdit(txn_type=CLAIM_DEF,
                                                                  field='some_field',
                                                                  old_value='old_value',
                                                                  new_value='new_value',
                                                                  is_owner=is_owner)])



from typing import Optional

from indy_common.authorize.auth_actions import AuthActionAdd, AuthActionEdit
from indy_common.authorize.auth_request_validator import WriteRequestValidator
from indy_common.constants import REVOC_REG_DEF, CRED_DEF_ID, REVOC_TYPE, TAG, ISSUANCE_BY_DEFAULT, ISSUANCE_ON_DEMAND
from indy_common.state.state_constants import MARKER_REVOC_DEF
from indy_node.server.revocation_strategy import RevokedStrategy, IssuedStrategy
from plenum.common.constants import DOMAIN_LEDGER_ID
from plenum.common.exceptions import InvalidClientRequest
from plenum.common.request import Request
from plenum.common.txn_util import get_seq_no, get_txn_time, get_from, get_payload_data

from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.handler_interfaces.write_request_handler import WriteRequestHandler
from plenum.server.request_handlers.utils import encode_state_value


class RevocRegDefHandler(WriteRequestHandler):
    revocation_strategy_map = {
        ISSUANCE_BY_DEFAULT: RevokedStrategy,
        ISSUANCE_ON_DEMAND: IssuedStrategy,
    }

    def __init__(self, database_manager: DatabaseManager,
                 write_req_validator: WriteRequestValidator):
        super().__init__(database_manager, REVOC_REG_DEF, DOMAIN_LEDGER_ID)
        self.write_req_validator = write_req_validator

    @staticmethod
    def get_revocation_strategy(typ):
        return RevocRegDefHandler.revocation_strategy_map.get(typ)

    def static_validation(self, request: Request):
        cred_def_id = request.operation.get(CRED_DEF_ID)
        tags = cred_def_id.split(":")
        if len(tags) != 4 and len(tags) != 5:
            raise InvalidClientRequest(request.identifier,
                                       request.reqId,
                                       "Format of {} field is not acceptable. "
                                       "Expected: 'did:marker:signature_type:schema_ref' or "
                                       "'did:marker:signature_type:schema_ref:tag'".format(CRED_DEF_ID))

    def additional_dynamic_validation(self, request: Request, req_pp_time: Optional[int]):
        self._validate_request_type(request)
        operation = request.operation
        cred_def_id = operation.get(CRED_DEF_ID)
        revoc_def_type = operation.get(REVOC_TYPE)
        revoc_def_tag = operation.get(TAG)
        assert cred_def_id
        assert revoc_def_tag
        assert revoc_def_type

        revoc_def_id = RevocRegDefHandler.make_state_path_for_revoc_def(request.identifier,
                                                                        cred_def_id,
                                                                        revoc_def_type,
                                                                        revoc_def_tag)
        revoc_def, _, _ = self.get_from_state(revoc_def_id)

        if revoc_def is None:
            self.write_req_validator.validate(request,
                                              [AuthActionAdd(txn_type=REVOC_REG_DEF,
                                                             field='*',
                                                             value='*')])
        else:
            self.write_req_validator.validate(request,
                                              [AuthActionEdit(txn_type=REVOC_REG_DEF,
                                                              field='*',
                                                              old_value='*',
                                                              new_value='*')])

        cred_def, _, _ = self.get_from_state(cred_def_id)
        if cred_def is None:
            raise InvalidClientRequest(request.identifier,
                                       request.reqId,
                                       "There is no any CRED_DEF by path: {}".format(cred_def_id))

    def gen_txn_id(self, txn):
        self._validate_txn_type(txn)
        path = RevocRegDefHandler.prepare_revoc_def_for_state(txn, path_only=True)
        return path.decode()

    def update_state(self, txn, prev_result, request, is_committed=False):
        self._validate_txn_type(txn)
        path, value_bytes = RevocRegDefHandler.prepare_revoc_def_for_state(txn)
        self.state.set(path, value_bytes)
        return txn

    @staticmethod
    def prepare_revoc_def_for_state(txn, path_only=False):
        author_did = get_from(txn)
        txn_data = get_payload_data(txn)
        cred_def_id = txn_data.get(CRED_DEF_ID)
        revoc_def_type = txn_data.get(REVOC_TYPE)
        revoc_def_tag = txn_data.get(TAG)
        assert author_did
        assert cred_def_id
        assert revoc_def_type
        assert revoc_def_tag
        path = RevocRegDefHandler.make_state_path_for_revoc_def(author_did,
                                                                cred_def_id,
                                                                revoc_def_type,
                                                                revoc_def_tag)
        if path_only:
            return path
        seq_no = get_seq_no(txn)
        txn_time = get_txn_time(txn)
        assert seq_no
        assert txn_time
        value_bytes = encode_state_value(txn_data, seq_no, txn_time)
        return path, value_bytes

    @staticmethod
    def make_state_path_for_revoc_def(authors_did, cred_def_id, revoc_def_type, revoc_def_tag) -> bytes:
        return "{DID}:{MARKER}:{CRED_DEF_ID}:{REVOC_DEF_TYPE}:{REVOC_DEF_TAG}" \
            .format(DID=authors_did,
                    MARKER=MARKER_REVOC_DEF,
                    CRED_DEF_ID=cred_def_id,
                    REVOC_DEF_TYPE=revoc_def_type,
                    REVOC_DEF_TAG=revoc_def_tag).encode()



import json
import copy

from indy_node.test.anon_creds.helper import build_get_revoc_reg_entry
from plenum.common.util import get_utc_epoch
from plenum.test.helper import sdk_send_and_check
from indy_common.constants import TIMESTAMP, REVOC_REG_DEF_ID, VALUE, ACCUM
from plenum.common.constants import TXN_TIME
from plenum.common.types import f


def test_send_get_revoc_reg_later_then_first_entry(looper,
                            txnPoolNodeSet,
                            sdk_pool_handle,
                            send_revoc_reg_entry_by_default,
                            sdk_wallet_steward):
    rev_entry_req, reg_reply = send_revoc_reg_entry_by_default
    get_revoc_reg = build_get_revoc_reg_entry(looper, sdk_wallet_steward)
    get_revoc_reg['operation'][REVOC_REG_DEF_ID] = rev_entry_req['operation'][REVOC_REG_DEF_ID]
    get_revoc_reg['operation'][TIMESTAMP] = get_utc_epoch() + 1000
    sdk_reply = sdk_send_and_check([json.dumps(get_revoc_reg)], looper, txnPoolNodeSet, sdk_pool_handle)
    reply = sdk_reply[0][1]
    assert rev_entry_req['operation'][REVOC_REG_DEF_ID] == reply['result'][REVOC_REG_DEF_ID]
    assert rev_entry_req['operation'][VALUE][ACCUM] == reply['result']['data'][VALUE][ACCUM]


def test_send_get_revoc_reg_earlier_then_first_entry(
        looper,
        txnPoolNodeSet,
        sdk_pool_handle,
        send_revoc_reg_entry_by_default,
        sdk_wallet_steward):
    rev_entry_req, reg_reply = send_revoc_reg_entry_by_default
    get_revoc_reg = build_get_revoc_reg_entry(looper, sdk_wallet_steward)
    get_revoc_reg['operation'][REVOC_REG_DEF_ID] = rev_entry_req['operation'][REVOC_REG_DEF_ID]
    get_revoc_reg['operation'][TIMESTAMP] = get_utc_epoch() - 1000

    sdk_reply = sdk_send_and_check([json.dumps(get_revoc_reg)], looper, txnPoolNodeSet, sdk_pool_handle)
    reply = sdk_reply[0][1]
    assert reply['result']['data'] is None
    assert reply['result'][TXN_TIME] is None
    assert reply['result'][f.SEQ_NO.nm] is None



import pytest

from indy_common.authorize.auth_actions import EDIT_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraint
from indy_node.test.auth_rule.helper import create_verkey_did, sdk_send_and_check_auth_rule_request
from plenum.common.constants import STEWARD_STRING, STEWARD, NYM, ROLE
from plenum.common.exceptions import RequestRejectedException
from plenum.test.helper import sdk_sign_and_submit_op, sdk_get_and_check_replies
from plenum.test.pool_transactions.helper import sdk_add_new_nym


def test_check_rule_for_edit_action_changing(looper,
                                             sdk_wallet_trustee,
                                             sdk_wallet_steward,
                                             sdk_pool_handle):
    wh, _ = sdk_wallet_trustee
    new_steward_did, new_steward_verkey = create_verkey_did(looper, wh)
    """Adding new steward for old auth rules"""
    sdk_add_new_nym(looper,
                    sdk_pool_handle,
                    sdk_wallet_trustee,
                    'newSteward1',
                    STEWARD_STRING,
                    dest=new_steward_did, verkey=new_steward_verkey)

    constraint = AuthConstraint(role=STEWARD,
                                sig_count=1)
    """Change role from steward to '' (blacklisting STEWARD) can only STEWARD"""
    sdk_send_and_check_auth_rule_request(looper, sdk_pool_handle, sdk_wallet_trustee,
                                         auth_action=EDIT_PREFIX,
                                         auth_type=NYM, field=ROLE, new_value='', old_value=STEWARD,
                                         constraint=constraint.as_dict)
    op = {'type': '1',
          'dest': new_steward_did,
          'role': None}
    with pytest.raises(RequestRejectedException, match="Not enough STEWARD signatures"):
        """Blacklisting new steward by TRUSTEE"""
        req = sdk_sign_and_submit_op(looper, sdk_pool_handle, sdk_wallet_trustee, op)
        sdk_get_and_check_replies(looper, [req])
    """We change auth rules and for now only steward can add new steward"""
    req = sdk_sign_and_submit_op(looper, sdk_pool_handle, sdk_wallet_steward, op)
    sdk_get_and_check_replies(looper, [req])



import json

import pytest
from libnacl.secret import SecretBox

from indy_common.constants import TXN_TYPE, ATTRIB, ENDPOINT
from indy_common.types import ClientAttribOperation
from plenum.common.constants import TARGET_NYM, RAW, ENC, HASH

validator = ClientAttribOperation()

VALID_TARGET_NYM = 'a' * 43
VALID_HASH = '6d4a333838d0ef96756cccC680AF2531075C512502Fb68c5503c63d93de859b3'
assert len(VALID_HASH) == 64


def test_attrib_with_enc_raw_hash_at_same_time_fails():
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        RAW: '{}',
        ENC: 'foo',
        HASH: VALID_HASH
    }
    with pytest.raises(TypeError) as ex_info:
        validator.validate(msg)
    ex_info.match(r"validation error \[ClientAttribOperation\]: only one field "
                  "from {}, {}, {} is expected"
                  "".format(RAW, ENC, HASH))


def test_attrib_without_enc_raw_hash_fails():
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM
    }
    with pytest.raises(TypeError) as ex_info:
        validator.validate(msg)
    ex_info.match(
        r"validation error \[ClientAttribOperation\]: missed fields - {}, {}, {}"
        "".format(
            RAW,
            ENC,
            HASH))


def test_attrib_with_raw_string_fails():
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        RAW: 'foo',
    }
    with pytest.raises(TypeError) as ex_info:
        validator.validate(msg)
    ex_info.match(r"validation error \[ClientAttribOperation\]: should be a "
                  r"valid JSON string \({}=foo\)".format(RAW))


def test_attrib_with_raw_empty_json_fails():
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        RAW: '{}',
    }
    with pytest.raises(TypeError) as ex_info:
        validator.validate(msg)
    ex_info.match(
        r"validation error \[ClientAttribOperation\]: should contain one attribute "
        r"\({}={{}}\)".format(RAW))


def test_attrib_with_raw_array_fails():
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        RAW: '[]',
    }
    with pytest.raises(TypeError) as ex_info:
        validator.validate(msg)
    ex_info.match(
        r"validation error \[ClientAttribOperation\]: should be a dict "
        r"\({}=<class 'list'>\)".format(RAW))


def test_attrib_with_raw_having_more_one_attrib_fails():
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        RAW: '{"attr1": "foo", "attr2": "bar"}',
    }
    with pytest.raises(TypeError) as ex_info:
        validator.validate(msg)
    ex_info.match(
        r"validation error \[ClientAttribOperation\]: should contain one attribute "
        r"\({}={{.*}}\)".format(RAW))


def test_attrib_with_raw_having_one_attrib_passes():
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        RAW: '{"attr1": "foo"}',
    }
    validator.validate(msg)


def test_attrib_with_raw_having_endpoint_equal_null_passes():
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        RAW: '{"endpoint": null}',
    }
    validator.validate(msg)


def test_attrib_with_raw_having_endpoint_ha_equal_null_passes():
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        RAW: '{"endpoint": {"ha": null}}',
    }
    validator.validate(msg)


def test_attrib_with_raw_having_endpoint_without_ha_passes():
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        RAW: '{"endpoint": {"foo": "bar"}}',
    }
    validator.validate(msg)


def test_attrib_with_raw_having_endpoint_ha_with_ip_address_only_fails():
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        RAW: '{"endpoint": {"ha": "8.8.8.8"}}',
    }
    with pytest.raises(TypeError) as ex_info:
        validator.validate(msg)
    ex_info.match(
        r"validation error \[ClientAttribOperation\]: invalid endpoint format ip_address:port "
        r"\({}={{'ha': '8.8.8.8'}}\)".format(ENDPOINT))


def test_attrib_with_raw_having_endpoint_ha_with_invalid_port_fails():
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        RAW: '{"endpoint": {"ha": "8.8.8.8:65536"}}',
    }
    with pytest.raises(TypeError) as ex_info:
        validator.validate(msg)
    ex_info.match(
        r"validation error \[ClientAttribOperation\]: invalid endpoint port "
        r"\(ha=8.8.8.8:65536\)")


def test_attrib_with_raw_having_endpoint_ha_with_invalid_ip_address_fails():
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        RAW: '{"endpoint": {"ha": "256.8.8.8:9700"}}',
    }
    with pytest.raises(TypeError) as ex_info:
        validator.validate(msg)
    ex_info.match(
        r"validation error \[ClientAttribOperation\]: invalid endpoint address "
        r"\(ha=256.8.8.8:9700\)")


def test_attrib_with_valid_hash_passes():
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        HASH: VALID_HASH
    }
    validator.validate(msg)


def test_attrib_with_shorter_hash_fails():
    invalid_hash = VALID_HASH[:-1]
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        HASH: invalid_hash
    }
    with pytest.raises(TypeError) as ex_info:
        validator.validate(msg)
    ex_info.match(r"validation error \[ClientAttribOperation\]: not a valid hash "
                  r"\(needs to be in hex too\) \({}={}\)".format(HASH, invalid_hash))


def test_attrib_with_longer_hash_fails():
    invalid_hash = VALID_HASH + 'a'
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        HASH: invalid_hash
    }
    with pytest.raises(TypeError) as ex_info:
        validator.validate(msg)
    ex_info.match(r"validation error \[ClientAttribOperation\]: not a valid hash "
                  r"\(needs to be in hex too\) \({}={}\)".format(HASH, invalid_hash))


def test_attrib_with_invalid_hash_fails():
    idx = 10
    invalid_hash = VALID_HASH[:idx] + 'X' + VALID_HASH[idx + 1:]
    assert len(invalid_hash) == 64
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        HASH: invalid_hash
    }
    with pytest.raises(TypeError) as ex_info:
        validator.validate(msg)
    ex_info.match(r"validation error \[ClientAttribOperation\]: not a valid hash "
                  r"\(needs to be in hex too\) \({}={}\)".format(HASH, invalid_hash))


def test_attrib_with_empty_hash_fails():
    empty_hash = ''
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        HASH: empty_hash
    }
    with pytest.raises(TypeError) as ex_info:
        validator.validate(msg)
    ex_info.match(r"validation error \[ClientAttribOperation\]: not a valid hash "
                  r"\(needs to be in hex too\) \({}={}\)".format(HASH, empty_hash))

    empty_hash = None
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        HASH: empty_hash
    }
    with pytest.raises(TypeError) as ex_info:
        validator.validate(msg)
    ex_info.match(r"validation error \[ClientAttribOperation\]: expected types "
                  r"'str', got 'NoneType' \({}={}\)".format(HASH, empty_hash))


def test_attrib_with_enc_passes():
    secretBox = SecretBox()
    enc_data = secretBox.encrypt(json.dumps({'name': 'Alice'}).encode()).hex()

    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        ENC: enc_data
    }
    validator.validate(msg)


def test_attrib_with_empty_enc_fails():
    empty_enc = ''
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        ENC: empty_enc
    }
    with pytest.raises(TypeError) as ex_info:
        validator.validate(msg)
    ex_info.match(r"validation error \[ClientAttribOperation\]: "
                  r"empty string \({}={}\)".format(ENC, empty_enc))

    empty_enc = None
    msg = {
        TXN_TYPE: ATTRIB,
        TARGET_NYM: VALID_TARGET_NYM,
        ENC: empty_enc
    }
    with pytest.raises(TypeError) as ex_info:
        validator.validate(msg)
    ex_info.match(r"validation error \[ClientAttribOperation\]: expected types "
                  r"'str', got 'NoneType' \({}={}\)".format(ENC, empty_enc))



import shutil

from plenum.common.constants import DOMAIN_LEDGER_ID
from plenum.common.util import randomString
from plenum.test.node_catchup.helper import ensure_all_nodes_have_same_data, \
    waitNodeDataEquality
from plenum.test.pool_transactions.helper import sdk_add_new_nym
from plenum.test.test_node import checkNodesConnected, ensure_node_disconnected
from indy_common.constants import ENDORSER_STRING
from indy_common.config_helper import NodeConfigHelper
from indy_node.test.helper import TestNode, sdk_add_raw_attribute

TestRunningTimeLimitSec = 200


def test_state_regenerated_from_ledger(looper,
                                       nodeSet, tconf, tdir,
                                       sdk_pool_handle,
                                       sdk_wallet_trustee,
                                       allPluginsPath):
    """
    Node loses its state database but recreates it from ledger after start.
    Checking ATTRIB txns too since they store some data off ledger too
    """
    endorsers = []
    for i in range(5):
        endorsers.append(sdk_add_new_nym(looper, sdk_pool_handle,
                                             sdk_wallet_trustee,
                                             'TA' + str(i),
                                             ENDORSER_STRING))
        sdk_add_raw_attribute(looper, sdk_pool_handle,
                              endorsers[-1],
                              randomString(6),
                              randomString(10))

    for wh in endorsers:
        for i in range(3):
            sdk_add_new_nym(looper, sdk_pool_handle,
                            wh, 'NP1' + str(i))

    ensure_all_nodes_have_same_data(looper, nodeSet)

    node_to_stop = nodeSet[-1]
    node_state = node_to_stop.states[DOMAIN_LEDGER_ID]
    assert not node_state.isEmpty
    state_db_path = node_state._kv.db_path
    node_to_stop.cleanupOnStopping = False
    node_to_stop.stop()
    looper.removeProdable(node_to_stop)
    ensure_node_disconnected(looper, node_to_stop, nodeSet[:-1])

    shutil.rmtree(state_db_path)

    config_helper = NodeConfigHelper(node_to_stop.name, tconf, chroot=tdir)
    restarted_node = TestNode(
        node_to_stop.name,
        config_helper=config_helper,
        config=tconf,
        pluginPaths=allPluginsPath,
        ha=node_to_stop.nodestack.ha,
        cliha=node_to_stop.clientstack.ha)
    looper.add(restarted_node)
    nodeSet[-1] = restarted_node

    looper.run(checkNodesConnected(nodeSet))
    # Need some time as `last_ordered_3PC` is compared too and that is
    # communicated through catchup
    waitNodeDataEquality(looper, restarted_node, *nodeSet[:-1])

    # Pool is still functional
    for wh in endorsers:
        sdk_add_new_nym(looper, sdk_pool_handle,
                        wh, 'NP--' + randomString(5))

    ensure_all_nodes_have_same_data(looper, nodeSet)



import copy
import json

import pytest

from indy_common.constants import RS_CONTENT, ENDORSER, RS_ENC_ALGORITHM
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_encoding_handler import \
    RichSchemaEncodingHandler
from indy_node.test.request_handlers.helper import add_to_idr
from indy_node.test.request_handlers.rich_schema.helper import rich_schema_encoding_request
from plenum.common.constants import TRUSTEE
from plenum.common.exceptions import InvalidClientRequest


@pytest.fixture()
def encoding_handler(db_manager, write_auth_req_validator):
    return RichSchemaEncodingHandler(db_manager, write_auth_req_validator)


@pytest.fixture()
def encoding_req(encoding_handler):
    req = rich_schema_encoding_request()
    add_to_idr(encoding_handler.database_manager.idr_cache, req.identifier, TRUSTEE)
    add_to_idr(encoding_handler.database_manager.idr_cache, req.endorser, ENDORSER)
    return req


@pytest.mark.parametrize('missing_field', ['input', 'output', 'algorithm', 'testVectors'])
@pytest.mark.parametrize('status', ['missing', 'empty', 'none'])
def test_static_validation_no_field(encoding_handler, encoding_req, missing_field, status):
    content = copy.deepcopy(json.loads(encoding_req.operation[RS_CONTENT]))
    if status == 'missing':
        content.pop(missing_field, None)
    elif status == 'empty':
        content[missing_field] = {}
    elif status == 'none':
        content[missing_field] = None
    encoding_req.operation[RS_CONTENT] = json.dumps(content)

    with pytest.raises(InvalidClientRequest,
                       match="'{}' must be set in 'content'".format(missing_field)):
        encoding_handler.static_validation(encoding_req)


@pytest.mark.parametrize('status', ['missing', 'empty', 'none'])
def test_static_validation_no_all_fields(encoding_handler, encoding_req, status):
    content = copy.deepcopy(json.loads(encoding_req.operation[RS_CONTENT]))
    if status == 'missing':
        content.pop('input', None)
        content.pop('output', None)
        content.pop('algorithm', None)
        content.pop('testVectors', None)
    elif status == 'empty':
        content['input'] = {}
        content['output'] = {}
        content['algorithm'] = {}
        content['testVectors'] = {}
    elif status == 'none':
        content['input'] = None
        content['output'] = None
        content['algorithm'] = None
        content['testVectors'] = None
    encoding_req.operation[RS_CONTENT] = json.dumps(content)

    with pytest.raises(InvalidClientRequest,
                       match="'input', 'output', 'algorithm', 'testVectors' must be set in 'content'"):
        encoding_handler.static_validation(encoding_req)


@pytest.mark.parametrize('missing_field', ['id', 'type'])
@pytest.mark.parametrize('input_output', ['input', 'output'])
@pytest.mark.parametrize('status', ['missing', 'empty', 'none'])
def test_static_validation_input_output(encoding_handler, encoding_req, missing_field, input_output, status):
    content = copy.deepcopy(json.loads(encoding_req.operation[RS_CONTENT]))
    if status == 'missing':
        content[input_output].pop(missing_field, None)
    elif status == 'empty':
        content[input_output][missing_field] = {}
    elif status == 'none':
        content[input_output][missing_field] = None
    encoding_req.operation[RS_CONTENT] = json.dumps(content)

    with pytest.raises(InvalidClientRequest,
                       match="'{}' must be set in '{}'".format(missing_field, input_output)):
        encoding_handler.static_validation(encoding_req)


@pytest.mark.parametrize('input_output', ['input', 'output'])
@pytest.mark.parametrize('status', ['empty', 'none'])
def test_static_validation_input_output_all_missing(encoding_handler, encoding_req, input_output,
                                                    status):
    content = copy.deepcopy(json.loads(encoding_req.operation[RS_CONTENT]))
    if status == 'empty':
        content[input_output]['id'] = {}
        content[input_output]['type'] = {}
    elif status == 'none':
        content[input_output]['id'] = None
        content[input_output]['type'] = None
    encoding_req.operation[RS_CONTENT] = json.dumps(content)

    with pytest.raises(InvalidClientRequest,
                       match="'id' and 'type' must be set in '{}'".format(input_output)):
        encoding_handler.static_validation(encoding_req)


@pytest.mark.parametrize('missing_field', ['description', 'documentation', 'implementation'])
@pytest.mark.parametrize('status', ['missing', 'empty', 'none'])
def test_static_validation_algorithm(encoding_handler, encoding_req, missing_field, status):
    content = copy.deepcopy(json.loads(encoding_req.operation[RS_CONTENT]))
    if status == 'missing':
        content[RS_ENC_ALGORITHM].pop(missing_field, None)
    elif status == 'empty':
        content[RS_ENC_ALGORITHM][missing_field] = {}
    elif status == 'none':
        content[RS_ENC_ALGORITHM][missing_field] = None
    encoding_req.operation[RS_CONTENT] = json.dumps(content)

    with pytest.raises(InvalidClientRequest,
                       match="'{}' must be set in '{}'".format(missing_field, RS_ENC_ALGORITHM)):
        encoding_handler.static_validation(encoding_req)


@pytest.mark.parametrize('status', ['empty', 'none'])
def test_static_validation_algorithm_all_missing(encoding_handler, encoding_req, status):
    content = copy.deepcopy(json.loads(encoding_req.operation[RS_CONTENT]))
    if status == 'empty':
        content[RS_ENC_ALGORITHM]['description'] = {}
        content[RS_ENC_ALGORITHM]['documentation'] = {}
        content[RS_ENC_ALGORITHM]['implementation'] = {}
    elif status == 'none':
        content[RS_ENC_ALGORITHM]['description'] = None
        content[RS_ENC_ALGORITHM]['documentation'] = None
        content[RS_ENC_ALGORITHM]['implementation'] = None
    encoding_req.operation[RS_CONTENT] = json.dumps(content)

    with pytest.raises(InvalidClientRequest,
                       match="'description', 'documentation', 'implementation' must be set in '{}'".format(
                           RS_ENC_ALGORITHM)):
        encoding_handler.static_validation(encoding_req)


def test_dynamic_validation_passes(encoding_handler, encoding_req):
    encoding_handler.dynamic_validation(encoding_req, 0)



from indy.ledger import build_pool_config_request

from plenum.test.helper import sdk_get_and_check_replies, sdk_sign_and_submit_req


def sdk_pool_config_sent(looper, sdk_pool_handle, sdk_wallet_trustee, pool_config_data):
    _, did = sdk_wallet_trustee
    req = looper.loop.run_until_complete(build_pool_config_request(
        did, pool_config_data['writes'], pool_config_data['force']))
    req = sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, req)
    return req



def sdk_ensure_pool_config_sent(looper, sdk_pool_handle, sdk_wallet_trustee, pool_config_data):
    _, did = sdk_wallet_trustee
    req = looper.loop.run_until_complete(build_pool_config_request(
        did, pool_config_data['writes'], pool_config_data['force']))
    req = sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_trustee, req)
    rep = sdk_get_and_check_replies(looper, [req])
    return rep


def check_pool_config_writable_set(nodes, writable):
    for node in nodes:
        assert node.poolCfg.isWritable() == writable



import json
import types

import pytest

from indy.did import create_and_store_my_did
from indy_node.persistence.idr_cache import IdrCache
from plenum.common.constants import VERKEY, DOMAIN_LEDGER_ID
from plenum.common.messages.node_messages import Commit
from plenum.common.util import randomString
from plenum.test.helper import sdk_get_and_check_replies
from indy_node.test.helper import sdk_rotate_verkey
from plenum.test.node_catchup.helper import waitNodeDataEquality
from plenum.test.delayers import icDelay
from plenum.test.pool_transactions.helper import sdk_add_new_nym, prepare_nym_request, \
    sdk_sign_and_send_prepared_request
from stp_core.common.log import getlogger
from stp_core.loop.eventually import eventually

logger = getlogger()


@pytest.fixture(scope="module")
def tconf(tconf, request):
    # Delaying performance checks since delaying 3PC messages in the test
    old_freq = tconf.PerfCheckFreq
    old_bt = tconf.Max3PCBatchWait
    old_bs = tconf.Max3PCBatchSize
    old_bf = tconf.Max3PCBatchesInFlight
    tconf.PerfCheckFreq = 50
    tconf.Max3PCBatchWait = .01
    tconf.Max3PCBatchSize = 1
    tconf.Max3PCBatchesInFlight = 5

    def reset():
        tconf.PerfCheckFreq = old_freq
        tconf.Max3PCBatchWait = old_bt
        tconf.Max3PCBatchSize = old_bs
        tconf.Max3PCBatchesInFlight = old_bf

    request.addfinalizer(reset)
    return tconf


@pytest.mark.skip(reason="get_req_handler is not supported yet")
def test_successive_batch_do_no_change_state(looper,
                                             tconf, nodeSet,
                                             sdk_pool_handle,
                                             sdk_wallet_trustee,
                                             monkeypatch):
    """
    Send 2 NYM txns in different batches such that the second batch does not
    change state so that state root remains same, but keep the identifier
    and reqId different. Make sure the first request is not ordered by the
    primary before PRE-PREPARE for the second is sent.
    Also check reject and commit
    :return:
    """

    # Disable view change during this test
    for n in nodeSet:
        n.nodeIbStasher.delay(icDelay())

    # Delay only first PRE-PREPARE
    delay_cm_duration = 10

    def delay_commits(wrappedMsg):
        msg, sender = wrappedMsg
        if isinstance(msg, Commit) and msg.instId == 0:
            return delay_cm_duration

    def check_verkey(i, vk):
        for node in nodeSet:
            data = node.idrCache.getNym(i, isCommitted=True)
            assert data[VERKEY] == vk

    def check_uncommitted(count):
        for node in nodeSet:
            assert len(node.idrCache.un_committed) == count

    for node in nodeSet:
        for rpl in node.replicas.values():
            monkeypatch.setattr(rpl, '_request_missing_three_phase_messages',
                                lambda *x, **y: None)

    wh, did = sdk_wallet_trustee
    seed = randomString(32)
    (new_did, verkey) = looper.loop.run_until_complete(
        create_and_store_my_did(wh, json.dumps({'seed': seed})))

    sdk_add_new_nym(looper, sdk_pool_handle,
                    sdk_wallet_trustee, dest=new_did,
                    verkey=verkey)

    for node in nodeSet:
        node.nodeIbStasher.delay(delay_commits)

    # Setting the same verkey thrice but in different batches with different
    # request ids
    for _ in range(3):
        verkey = sdk_rotate_verkey(looper, sdk_pool_handle,
                                   wh, new_did, new_did)
        logger.debug('{} rotates his key to {}'.
                     format(new_did, verkey))

    # Number of uncommitted entries is 0
    looper.run(eventually(check_uncommitted, 0))

    check_verkey(new_did, verkey)

    # Setting the verkey to `x`, then `y` and then back to `x` but in different
    # batches with different request ids. The idea is to change
    # state root to `t` then `t'` and then back to `t` and observe that no
    # errors are encountered

    seed = randomString(32)
    (new_client_did, verkey) = looper.loop.run_until_complete(
        create_and_store_my_did(wh, json.dumps({'seed': seed})))
    sdk_add_new_nym(looper, sdk_pool_handle,
                    sdk_wallet_trustee, dest=new_client_did,
                    verkey=verkey)

    x_seed = randomString(32)
    verkey = sdk_rotate_verkey(looper, sdk_pool_handle, wh,
                               new_client_did, new_client_did)
    logger.debug('{} rotates his key to {}'.
                 format(new_client_did, verkey))

    y_seed = randomString(32)
    sdk_rotate_verkey(looper, sdk_pool_handle, wh,
                      new_client_did, new_client_did)
    logger.debug('{} rotates his key to {}'.
                 format(new_client_did, verkey))

    verkey = sdk_rotate_verkey(looper, sdk_pool_handle, wh,
                      new_client_did, new_client_did)
    logger.debug('{} rotates his key to {}'.
                 format(new_client_did, verkey))

    # Number of uncommitted entries is 0
    looper.run(eventually(check_uncommitted, 0))

    check_verkey(new_client_did, verkey)
    monkeypatch.undo()

    # Delay COMMITs so that IdrCache can be checked for correct
    # number of entries

    uncommitteds = {}
    methods = {}
    for node in nodeSet:
        cache = node.idrCache  # type: IdrCache
        uncommitteds[cache._name] = []
        # Since the post batch creation handler is registered (added to a list),
        # find it and patch it
        dh = node.get_req_handler(DOMAIN_LEDGER_ID)
        for i, handler in enumerate(dh.post_batch_creation_handlers):
            # Find the cache's post create handler, not hardcoding names of
            # class or functions as they can change with refactoring.
            if handler.__func__.__qualname__ == '{}.{}'.format(cache.__class__.__name__,
                                                               cache.currentBatchCreated.__name__):
                cre = dh.post_batch_creation_handlers[i]
                break
        com = cache.onBatchCommitted
        methods[cache._name] = (cre, com)

        # Patch methods to record and check roots after commit

        def patched_cre(self, stateRoot, ppTime):
            uncommitteds[self._name].append(stateRoot)
            return methods[self._name][0](stateRoot, ppTime)

        def patched_com(self, stateRoot):
            assert uncommitteds[self._name][0] == stateRoot
            rv = methods[self._name][1](stateRoot)
            uncommitteds[self._name] = uncommitteds[self._name][1:]
            return rv

        dh.post_batch_creation_handlers[i] = types.MethodType(patched_cre, cache)
        cache.onBatchCommitted = types.MethodType(patched_com, cache)

    # Set verkey of multiple identities
    more = 5
    keys = {}
    reqs = []
    for _ in range(more):
        seed = randomString(32)
        nym_request, new_did = looper.loop.run_until_complete(
            prepare_nym_request(sdk_wallet_trustee, seed,
                                None, None))
        keys[new_did] = json.loads(nym_request)['operation']['verkey']
        reqs.append(
            sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                               sdk_pool_handle, nym_request))
        looper.runFor(tconf.Max3PCBatchWait + 0.1)

    # Correct number of uncommitted entries
    looper.run(eventually(check_uncommitted, more, retryWait=1))

    sdk_get_and_check_replies(looper, reqs)

    # Number of uncommitted entries is 0
    looper.run(eventually(check_uncommitted, 0))

    # The verkeys are correct
    for i, v in keys.items():
        check_verkey(i, v)

    waitNodeDataEquality(looper, nodeSet[0], *nodeSet[1:])

    for _ in range(3):
        seed = randomString(32)
        nym_request, new_did = looper.loop.run_until_complete(
            prepare_nym_request(sdk_wallet_trustee, seed,
                                None, None))
        reqs.append(
            sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee,
                                               sdk_pool_handle, nym_request))
        looper.runFor(tconf.Max3PCBatchWait + 0.1)

    # Correct number of uncommitted entries
    looper.run(eventually(check_uncommitted, 3, retryWait=1))

    # Check batch reject
    for node in nodeSet:
        cache = node.idrCache
        initial = cache.un_committed
        cache.batchRejected()
        # After reject, last entry is removed
        assert cache.un_committed == initial[:-1]
        root = cache.un_committed[0][0]
        cache.onBatchCommitted(root)
        # Calling commit with same root results in Assertion error
        with pytest.raises(AssertionError):
            cache.onBatchCommitted(root)



from indy_node.test.api.helper import validate_write_reply, validate_schema_txn, sdk_write_schema


def test_schema_reply_is_valid(looper, sdk_pool_handle, sdk_wallet_steward):
    _, reply = sdk_write_schema(looper, sdk_pool_handle, sdk_wallet_steward)
    validate_write_reply(reply)
    validate_schema_txn(reply['result']['txn'])



from indy_node.test.did.conftest import nym_get
from indy_node.test.helper import sdk_rotate_verkey


def testAddDidWithVerkey(nym_abbrv_vk):
    pass


def testRetrieveAbbrvVerkey(looper, tconf, nodeSet, sdk_pool_handle, sdk_wallet_trustee, nym_abbrv_vk):
    nwh, ndid, nvk = nym_abbrv_vk
    resp_data = nym_get(looper, sdk_pool_handle, sdk_wallet_trustee, ndid)
    assert ndid == resp_data[0]
    assert nvk == resp_data[1]


def testChangeVerkeyToNewVerkey(looper, tconf, nodeSet, sdk_pool_handle, nym_abbrv_vk):
    wh, did, nvk = nym_abbrv_vk
    new_verkey = sdk_rotate_verkey(looper, sdk_pool_handle, wh, did, did)
    assert nvk != new_verkey


def testRetrieveChangedVerkey(looper, tconf, nodeSet, sdk_pool_handle, sdk_wallet_trustee, nym_abbrv_vk):
    wh, did, vk = nym_abbrv_vk
    new_vk = sdk_rotate_verkey(looper, sdk_pool_handle, wh, did, did)
    resp_data = nym_get(looper, sdk_pool_handle, sdk_wallet_trustee, did)
    assert did == resp_data[0]
    assert vk != resp_data[1]
    assert new_vk == resp_data[1]


def testVerifySigWithChangedVerkey(looper, tconf, nodeSet, sdk_pool_handle, nym_abbrv_vk):
    wh, did, vk = nym_abbrv_vk
    new_vk = sdk_rotate_verkey(looper, sdk_pool_handle, wh, did, did)
    # check sign by getting nym from ledger - if succ then sign is ok
    resp_data = nym_get(looper, sdk_pool_handle, (wh, did), did)
    assert did == resp_data[0]
    assert vk != resp_data[1]
    assert new_vk == resp_data[1]



import multiprocessing
import os
import functools
import shutil

from stp_core.loop.eventually import eventually
from stp_core.common.log import getlogger

from indy_common.version import src_version_cls
from indy_node.server.upgrader import Upgrader
from indy_node.utils.node_control_utils import NodeControlUtil, DebianVersion

from indy_node.test.upgrade.helper import (
    NodeControlToolExecutor as NCT,
    sendUpgradeMessage,
    nodeControlGeneralMonkeypatching,
    bumpedVersion
)

m = multiprocessing.Manager()
# TODO why do we expect that
whitelist = ['Unexpected error in _upgrade test']
logger = getlogger()


def testNodeControlRemovesBackups(monkeypatch, tdir, looper, tconf):
    version = bumpedVersion()
    stdout = 'teststdout'
    curr_src_ver = Upgrader.get_src_version()
    backupsWereRemoved = m.Value('b', False)

    def testRemoveOldBackups(tool):
        assert len(tool._get_backups()) == (tool.backup_num + 1)
        #looper = Looper(debug=True)
        tool._remove_old_backups_test()
        backupsWereRemoved.value = True

    def transform(tool):
        nodeControlGeneralMonkeypatching(tool, monkeypatch, tdir, stdout)
        tool._remove_old_backups_test = tool._remove_old_backups
        monkeypatch.setattr(tool, '_remove_old_backups',
                            functools.partial(testRemoveOldBackups, tool))

    def checkOldBackupsRemoved():
        assert backupsWereRemoved.value

    def check_backups_files_exists():
        assert os.path.exists('{}.{}'.format(
            nct.tool._backup_name(curr_src_ver.release), nct.tool.backup_format))

    monkeypatch.setattr(
        NodeControlUtil, 'get_latest_pkg_version',
        lambda *x, **y: DebianVersion(
            version, upstream_cls=src_version_cls())
    )

    nct = NCT(backup_dir=tdir, backup_target=tdir, transform=transform)
    try:
        assert len(nct.tool._get_backups()) == 0
        for i in range(nct.tool.backup_num):
            file = os.path.join(nct.tool.base_dir, '{}{}'.format(
                nct.tool.backup_name_prefix, i))
            with open(file, 'w') as f:
                f.write('random')
        assert len(nct.tool._get_backups()) == nct.tool.backup_num
        sendUpgradeMessage(version)
        looper.run(eventually(checkOldBackupsRemoved))
        looper.run(eventually(check_backups_files_exists))
        assert len(nct.tool._get_backups()) == nct.tool.backup_num
    finally:
        clean_dir(nct.tool.base_dir)
        nct.stop()

def clean_dir(dir):
    for file in os.listdir(dir):
        file = os.path.join(dir, file)
        if os.path.isfile(file):
            os.remove(file)
        if os.path.isdir(file):
            shutil.rmtree(file, ignore_errors=True)



from collections import OrderedDict

from indy_common.types import ClientSchemaOperation, SchemaField
from plenum.common.messages.fields import ConstantField, VersionField, IterableField, LimitedLengthStringField

EXPECTED_ORDERED_FIELDS_SCHEMA = OrderedDict([
    ("name", LimitedLengthStringField),
    ("version", VersionField),
    ("attr_names", IterableField),
])


def test_has_expected_fields_s():
    actual_field_names = OrderedDict(SchemaField.schema).keys()
    assert actual_field_names == EXPECTED_ORDERED_FIELDS_SCHEMA.keys()


def test_has_expected_validators_s():
    schema = dict(SchemaField.schema)
    for field, validator in EXPECTED_ORDERED_FIELDS_SCHEMA.items():
        assert isinstance(schema[field], validator)


EXPECTED_ORDERED_FIELDS = OrderedDict([
    ("type", ConstantField),
    ("data", SchemaField),
])


def test_has_expected_fields():
    actual_field_names = OrderedDict(ClientSchemaOperation.schema).keys()
    assert actual_field_names == EXPECTED_ORDERED_FIELDS.keys()


def test_has_expected_validators():
    schema = dict(ClientSchemaOperation.schema)
    for field, validator in EXPECTED_ORDERED_FIELDS.items():
        assert isinstance(schema[field], validator)



import pytest as pytest

from common.serializers.serialization import domain_state_serializer, config_state_serializer
from indy_node.server.request_handlers.config_req_handlers.txn_author_agreement_handler import TxnAuthorAgreementHandler
from indy_node.test.request_handlers.helper import get_exception, add_to_idr
from plenum.common.constants import ROLE, STEWARD, NYM, TARGET_NYM, TXN_TYPE, TXN_AUTHOR_AGREEMENT, \
    TXN_AUTHOR_AGREEMENT_TEXT, TXN_AUTHOR_AGREEMENT_VERSION, TRUSTEE, DOMAIN_LEDGER_ID
from plenum.common.exceptions import UnauthorizedClientRequest, InvalidClientRequest
from plenum.common.request import Request
from plenum.common.txn_util import get_payload_data, reqToTxn, get_reply_nym
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.static_taa_helper import StaticTAAHelper
from plenum.server.request_handlers.utils import get_nym_details, get_role, is_steward, nym_to_state_key, \
    encode_state_value
from plenum.test.testing_utils import FakeSomething


@pytest.fixture(scope="module")
def txn_author_agreement_handler(db_manager, write_auth_req_validator):
    handler = TxnAuthorAgreementHandler(db_manager, write_auth_req_validator)
    return handler


@pytest.fixture(scope="module")
def set_aml(txn_author_agreement_handler):
    txn_author_agreement_handler.state.set(StaticTAAHelper.state_path_taa_aml_latest(),
                                           encode_state_value("value", "seqNo", "txnTime",
                                                              serializer=config_state_serializer))


def test_dynamic_validation_without_permission(taa_request,
                                               txn_author_agreement_handler: TxnAuthorAgreementHandler,
                                               creator,
                                               set_aml):
    add_to_idr(txn_author_agreement_handler.database_manager.idr_cache, creator, STEWARD)
    with pytest.raises(UnauthorizedClientRequest, match="Not enough TRUSTEE signatures"):
        txn_author_agreement_handler.dynamic_validation(taa_request, 0)


def test_dynamic_validation(taa_request,
                            txn_author_agreement_handler: TxnAuthorAgreementHandler,
                            creator,
                            set_aml):
    add_to_idr(txn_author_agreement_handler.database_manager.idr_cache, creator, TRUSTEE)
    txn_author_agreement_handler.dynamic_validation(taa_request, 0)




import json
import copy

from indy_node.test.anon_creds.helper import build_get_revoc_reg_delta
from plenum.common.txn_util import get_txn_time, get_payload_data
from plenum.common.util import get_utc_epoch
from plenum.test.helper import sdk_send_and_check, sdk_sign_request_from_dict
from indy_common.constants import REVOC_REG_DEF_ID, VALUE, FROM, TO, ISSUED, \
    REVOKED, PREV_ACCUM, ACCUM_FROM, ACCUM_TO, STATE_PROOF_FROM, ACCUM, REVOC_TYPE
from plenum.common.constants import TXN_TIME, DATA
from plenum.common.types import f
from plenum.common.util import randomString


def test_send_with_wrong_rev_reg_id_default(looper, txnPoolNodeSet,
                                            sdk_pool_handle,
                                            send_revoc_reg_entry_by_default,
                                            sdk_wallet_steward):
    get_revoc_reg_delta = build_get_revoc_reg_delta(looper, sdk_wallet_steward)
    del get_revoc_reg_delta['operation'][FROM]
    get_revoc_reg_delta['operation'][REVOC_REG_DEF_ID] = randomString(30)
    get_revoc_reg_delta['operation'][TO] = get_utc_epoch() + 1000
    sdk_reply = sdk_send_and_check([json.dumps(get_revoc_reg_delta)], looper,
                                   txnPoolNodeSet, sdk_pool_handle)
    reply = sdk_reply[0][1]
    assert DATA in reply['result']
    assert reply['result'][DATA] is None


def test_send_with_only_to_by_default(looper,
                            txnPoolNodeSet,
                            sdk_pool_handle,
                            send_revoc_reg_entry_by_default,
                            sdk_wallet_steward):
    rev_entry_req, reg_reply = send_revoc_reg_entry_by_default
    get_revoc_reg_delta = build_get_revoc_reg_delta(looper, sdk_wallet_steward)
    del get_revoc_reg_delta['operation'][FROM]
    get_revoc_reg_delta['operation'][REVOC_REG_DEF_ID] = rev_entry_req['operation'][REVOC_REG_DEF_ID]
    get_revoc_reg_delta['operation'][TO] = get_utc_epoch() + 1000
    sdk_reply = sdk_send_and_check([json.dumps(get_revoc_reg_delta)], looper, txnPoolNodeSet, sdk_pool_handle)
    reply = sdk_reply[0][1]
    assert rev_entry_req['operation'][REVOC_REG_DEF_ID] == reply['result'][REVOC_REG_DEF_ID]
    assert rev_entry_req['operation'][VALUE][ACCUM] == reply['result'][DATA][VALUE][ACCUM_TO][VALUE][ACCUM]
    assert rev_entry_req['operation'][VALUE][REVOKED] == reply['result'][DATA][VALUE][REVOKED]


def test_send_earlier_then_first_entry_by_default(
        looper,
        txnPoolNodeSet,
        sdk_pool_handle,
        send_revoc_reg_entry_by_default,
        sdk_wallet_steward):
    rev_entry_req, reg_reply = send_revoc_reg_entry_by_default
    get_revoc_reg_delta = build_get_revoc_reg_delta(looper, sdk_wallet_steward)
    del get_revoc_reg_delta['operation'][FROM]
    get_revoc_reg_delta['operation'][REVOC_REG_DEF_ID] = rev_entry_req['operation'][REVOC_REG_DEF_ID]
    get_revoc_reg_delta['operation'][TO] = get_utc_epoch() - 1000
    sdk_reply = sdk_send_and_check([json.dumps(get_revoc_reg_delta)], looper, txnPoolNodeSet, sdk_pool_handle)
    reply = sdk_reply[0][1]
    assert DATA in reply['result']
    assert reply['result'][DATA] is None
    assert reply['result'][f.SEQ_NO.nm] is None
    assert reply['result'][TXN_TIME] is None


def test_send_with_from_by_default(looper,
        txnPoolNodeSet,
        sdk_pool_handle,
        sdk_wallet_steward,
        send_revoc_reg_entry_by_default):
    # We save timestamp of state changes.
    # looper and txnPoolNodeSet has "module", therefore,
    # when we send request with FROM section, it's not a clean situation
    looper.runFor(3)
    # Assume, that send_revoc_reg_entry_by_default will add into revoked [1,2,3,4,5]
    rev_reg_req1, rev_reg_reply1 = send_revoc_reg_entry_by_default
    rev_reg_req1['operation'][VALUE][REVOKED] = []
    # Issue [1,2,3], Revoked now must be [4,5]
    rev_reg_req1['operation'][VALUE][ISSUED] = [1, 2, 3]
    rev_reg_req1['operation'][VALUE][PREV_ACCUM] = rev_reg_req1['operation'][VALUE][ACCUM]
    rev_reg_req1['operation'][VALUE][ACCUM] = randomString(10)
    rev_reg_req2, rev_reg_reply2 = sdk_send_and_check(
        [json.dumps(sdk_sign_request_from_dict(
            looper, sdk_wallet_steward, rev_reg_req1['operation']))],
        looper,
        txnPoolNodeSet,
        sdk_pool_handle)[0]
    # Revoke [10, 11]
    rev_reg_req2['operation'][VALUE][REVOKED] = [10, 11]
    rev_reg_req2['operation'][VALUE][ISSUED] = []
    rev_reg_req2['operation'][VALUE][PREV_ACCUM] = rev_reg_req2['operation'][VALUE][ACCUM]
    rev_reg_req2['operation'][VALUE][ACCUM] = randomString(10)
    rev_reg_req3, rev_reg_reply3 = sdk_send_and_check(
        [json.dumps(sdk_sign_request_from_dict(
            looper, sdk_wallet_steward, rev_reg_req2['operation']))],
        looper,
        txnPoolNodeSet,
        sdk_pool_handle)[0]
    reg_delta_req = build_get_revoc_reg_delta(looper, sdk_wallet_steward)
    reg_delta_req['operation'][REVOC_REG_DEF_ID] = rev_reg_req1['operation'][REVOC_REG_DEF_ID]
    reg_delta_req['operation'][FROM] = get_txn_time(rev_reg_reply1['result'])
    reg_delta_req['operation'][TO] = get_txn_time(rev_reg_reply3['result']) + 1000
    get_reply = sdk_send_and_check([json.dumps(reg_delta_req)], looper, txnPoolNodeSet, sdk_pool_handle)[0][1]
    assert get_reply['result'][DATA][STATE_PROOF_FROM]
    assert get_reply['result'][DATA][VALUE][ISSUED] == [1, 2, 3]
    assert get_reply['result'][DATA][VALUE][REVOKED] == [10, 11]
    assert get_reply['result'][DATA][VALUE][ACCUM_TO][VALUE][ACCUM] == rev_reg_req3['operation'][VALUE][ACCUM]
    assert get_reply['result'][DATA][VALUE][ACCUM_FROM][VALUE][ACCUM] == \
           get_payload_data(rev_reg_reply1['result'])[VALUE][ACCUM]



import pytest

from indy_common.constants import ENDORSER
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_handler import RichSchemaHandler
from indy_node.test.request_handlers.helper import add_to_idr
from indy_node.test.request_handlers.rich_schema.helper import rich_schema_request
from plenum.common.constants import TRUSTEE


@pytest.fixture()
def rich_schema_handler(db_manager, write_auth_req_validator):
    return RichSchemaHandler(db_manager, write_auth_req_validator)


@pytest.fixture()
def rich_schema_req(rich_schema_handler):
    req = rich_schema_request()
    add_to_idr(rich_schema_handler.database_manager.idr_cache, req.identifier, TRUSTEE)
    add_to_idr(rich_schema_handler.database_manager.idr_cache, req.endorser, ENDORSER)
    return req


def test_schema_dynamic_validation_passes(rich_schema_handler, rich_schema_req):
    rich_schema_handler.dynamic_validation(rich_schema_req, 0)



from common.serializers.serialization import domain_state_serializer
from indy_common.authorize.auth_constraints import ConstraintsSerializer
from indy_common.authorize.auth_map import auth_map
from indy_common.authorize.auth_request_validator import WriteRequestValidator
from indy_common.constants import CONFIG_LEDGER_ID
from indy_node.persistence.attribute_store import AttributeStore
from indy_node.persistence.idr_cache import IdrCache
from indy_node.server.pool_config import PoolConfig
from indy_node.server.request_handlers.action_req_handlers.pool_restart_handler import PoolRestartHandler
from indy_node.server.request_handlers.action_req_handlers.validator_info_handler import ValidatorInfoHandler
from indy_node.server.request_handlers.config_batch_handler import ConfigBatchHandler
from indy_node.server.request_handlers.config_req_handlers.auth_rule.auth_rule_handler import AuthRuleHandler
from indy_node.server.request_handlers.config_req_handlers.auth_rule.auth_rule_handler_1_9_1 import AuthRuleHandler191
from indy_node.server.request_handlers.config_req_handlers.auth_rule.auth_rules_handler import AuthRulesHandler
from indy_node.server.request_handlers.config_req_handlers.ledgers_freeze_handler import LedgersFreezeHandler
from indy_node.server.request_handlers.config_req_handlers.node_upgrade_handler import NodeUpgradeHandler
from indy_node.server.request_handlers.config_req_handlers.pool_config_handler import PoolConfigHandler
from indy_node.server.request_handlers.config_req_handlers.pool_upgrade_handler import PoolUpgradeHandler
from indy_node.server.request_handlers.config_req_handlers.txn_author_agreement_aml_handler import \
    TxnAuthorAgreementAmlHandler
from indy_node.server.request_handlers.config_req_handlers.txn_author_agreement_disable_handler import \
    TxnAuthorAgreementDisableHandler
from indy_node.server.request_handlers.config_req_handlers.txn_author_agreement_handler import TxnAuthorAgreementHandler
from indy_node.server.request_handlers.config_req_handlers.txn_author_agreement_handler_v1 import \
    TxnAuthorAgreementHandlerV1
from indy_node.server.request_handlers.domain_req_handlers.attribute_handler import AttributeHandler
from indy_node.server.request_handlers.domain_req_handlers.claim_def_handler import ClaimDefHandler
from indy_node.server.request_handlers.domain_req_handlers.idr_cache_nym_handler import IdrCacheNymHandler
from indy_node.server.request_handlers.domain_req_handlers.nym_handler import NymHandler
from indy_node.server.request_handlers.domain_req_handlers.revoc_reg_def_handler import RevocRegDefHandler
from indy_node.server.request_handlers.domain_req_handlers.revoc_reg_entry_handler import RevocRegEntryHandler
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.json_ld_context_handler import \
    JsonLdContextHandler
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_cred_def_handler import \
    RichSchemaCredDefHandler
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_encoding_handler import \
    RichSchemaEncodingHandler
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_handler import RichSchemaHandler
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_mapping_handler import \
    RichSchemaMappingHandler
from indy_node.server.request_handlers.domain_req_handlers.rich_schema.rich_schema_pres_def_handler import \
    RichSchemaPresDefHandler
from indy_node.server.request_handlers.domain_req_handlers.schema_handler import SchemaHandler
from indy_node.server.request_handlers.idr_cache_batch_handler import IdrCacheBatchHandler
from indy_node.server.request_handlers.pool_req_handlers.node_handler import NodeHandler
from indy_node.server.request_handlers.read_req_handlers.get_attribute_handler import GetAttributeHandler
from indy_node.server.request_handlers.read_req_handlers.get_auth_rule_handler import GetAuthRuleHandler
from indy_node.server.request_handlers.read_req_handlers.get_claim_def_handler import GetClaimDefHandler
from plenum.server.request_handlers.ledgers_freeze.get_frozen_ledgers_handler import GetFrozenLedgersHandler
from indy_node.server.request_handlers.read_req_handlers.get_nym_handler import GetNymHandler
from indy_node.server.request_handlers.read_req_handlers.get_revoc_reg_def_handler import GetRevocRegDefHandler
from indy_node.server.request_handlers.read_req_handlers.get_revoc_reg_delta_handler import GetRevocRegDeltaHandler
from indy_node.server.request_handlers.read_req_handlers.get_revoc_reg_handler import GetRevocRegHandler
from indy_node.server.request_handlers.read_req_handlers.rich_schema.get_rich_schema_object_by_id_handler import \
    GetRichSchemaObjectByIdHandler
from indy_node.server.request_handlers.read_req_handlers.rich_schema.get_rich_schema_object_by_metadata_handler import \
    GetRichSchemaObjectByMetadataHandler
from indy_node.server.request_handlers.read_req_handlers.get_schema_handler import GetSchemaHandler
from indy_node.server.restarter import Restarter
from indy_node.server.upgrader import Upgrader
from plenum.common.constants import IDR_CACHE_LABEL, ATTRIB_LABEL
from plenum.server.node_bootstrap import NodeBootstrap as PNodeBootstrap
from plenum.server.request_handlers.get_txn_author_agreement_aml_handler import GetTxnAuthorAgreementAmlHandler
from plenum.server.request_handlers.get_txn_author_agreement_handler import GetTxnAuthorAgreementHandler
from storage.helper import initKeyValueStorage


class NodeBootstrap(PNodeBootstrap):

    def init_idr_cache_storage(self):
        idr_cache = IdrCache(self.node.name,
                             initKeyValueStorage(self.node.config.idrCacheStorage,
                                                 self.node.dataLocation,
                                                 self.node.config.idrCacheDbName,
                                                 db_config=self.node.config.db_idr_cache_db_config)
                             )
        self.node.db_manager.register_new_store(IDR_CACHE_LABEL, idr_cache)

    def init_attribute_store(self):
        return AttributeStore(
            initKeyValueStorage(
                self.node.config.attrStorage,
                self.node.dataLocation,
                self.node.config.attrDbName,
                db_config=self.node.config.db_attr_db_config)
        )

    def init_attribute_storage(self):
        # ToDo: refactor this on pluggable handlers integration phase
        attr_store = self.init_attribute_store()
        self.node.db_manager.register_new_store(ATTRIB_LABEL, attr_store)

    def _init_storages(self, domain_storage=None):
        super()._init_storages(domain_storage)
        self.init_idr_cache_storage()
        self.init_attribute_storage()

    def _register_pool_req_handlers(self):
        node_handler = NodeHandler(self.node.db_manager,
                                   self.node.bls_bft.bls_crypto_verifier,
                                   self.node.write_req_validator)
        self.node.write_manager.register_req_handler(node_handler)

    def _register_domain_req_handlers(self):
        # Read handlers
        get_nym_handler = GetNymHandler(node=self.node, database_manager=self.node.db_manager)
        get_attribute_handler = GetAttributeHandler(node=self.node, database_manager=self.node.db_manager)
        get_schema_handler = GetSchemaHandler(database_manager=self.node.db_manager)
        get_claim_def_handler = GetClaimDefHandler(database_manager=self.node.db_manager)
        get_revoc_reg_def_handler = GetRevocRegDefHandler(database_manager=self.node.db_manager)
        get_revoc_reg_handler = GetRevocRegHandler(database_manager=self.node.db_manager)
        get_revoc_reg_delta_handler = GetRevocRegDeltaHandler(database_manager=self.node.db_manager,
                                                              get_revocation_strategy=RevocRegDefHandler.get_revocation_strategy)
        # Write handlers
        nym_handler = NymHandler(config=self.node.config,
                                 database_manager=self.node.db_manager,
                                 write_req_validator=self.node.write_req_validator)
        attrib_handler = AttributeHandler(database_manager=self.node.db_manager,
                                          write_req_validator=self.node.write_req_validator)
        schema_handler = SchemaHandler(database_manager=self.node.db_manager,
                                       write_req_validator=self.node.write_req_validator)
        claim_def_handler = ClaimDefHandler(database_manager=self.node.db_manager,
                                            write_req_validator=self.node.write_req_validator)
        revoc_reg_def_handler = RevocRegDefHandler(database_manager=self.node.db_manager,
                                                   write_req_validator=self.node.write_req_validator)
        revoc_reg_entry_handler = RevocRegEntryHandler(database_manager=self.node.db_manager,
                                                       write_req_validator=self.node.write_req_validator,
                                                       get_revocation_strategy=RevocRegDefHandler.get_revocation_strategy)
        json_ld_context_handler = JsonLdContextHandler(database_manager=self.node.db_manager,
                                                       write_req_validator=self.node.write_req_validator)
        rich_schema_handler = RichSchemaHandler(database_manager=self.node.db_manager,
                                                write_req_validator=self.node.write_req_validator)
        rich_schema_encoding_handler = RichSchemaEncodingHandler(database_manager=self.node.db_manager,
                                                                 write_req_validator=self.node.write_req_validator)
        rich_schema_mapping_handler = RichSchemaMappingHandler(database_manager=self.node.db_manager,
                                                               write_req_validator=self.node.write_req_validator)
        rich_schema_cred_def_handler = RichSchemaCredDefHandler(database_manager=self.node.db_manager,
                                                                write_req_validator=self.node.write_req_validator)
        rich_schema_pres_def_handler = RichSchemaPresDefHandler(database_manager=self.node.db_manager,
                                                                write_req_validator=self.node.write_req_validator)
        get_rich_schema_obj_by_id_handler = GetRichSchemaObjectByIdHandler(database_manager=self.node.db_manager)
        get_rich_schema_obj_by_metadata_handler = GetRichSchemaObjectByMetadataHandler(
            database_manager=self.node.db_manager)

        # Register write handlers
        self.node.write_manager.register_req_handler(nym_handler)
        self.node.write_manager.register_req_handler(attrib_handler)
        self.node.write_manager.register_req_handler(schema_handler)
        self.node.write_manager.register_req_handler(claim_def_handler)
        self.node.write_manager.register_req_handler(revoc_reg_def_handler)
        self.node.write_manager.register_req_handler(revoc_reg_entry_handler)
        self.node.write_manager.register_req_handler(json_ld_context_handler)
        self.node.write_manager.register_req_handler(rich_schema_handler)
        self.node.write_manager.register_req_handler(rich_schema_encoding_handler)
        self.node.write_manager.register_req_handler(rich_schema_mapping_handler)
        self.node.write_manager.register_req_handler(rich_schema_cred_def_handler)
        self.node.write_manager.register_req_handler(rich_schema_pres_def_handler)
        # Additional handler for idCache
        self.register_idr_cache_nym_handler()
        # Register read handlers
        self.node.read_manager.register_req_handler(get_nym_handler)
        self.node.read_manager.register_req_handler(get_attribute_handler)
        self.node.read_manager.register_req_handler(get_schema_handler)
        self.node.read_manager.register_req_handler(get_claim_def_handler)
        self.node.read_manager.register_req_handler(get_revoc_reg_def_handler)
        self.node.read_manager.register_req_handler(get_revoc_reg_handler)
        self.node.read_manager.register_req_handler(get_revoc_reg_delta_handler)
        self.node.read_manager.register_req_handler(get_rich_schema_obj_by_id_handler)
        self.node.read_manager.register_req_handler(get_rich_schema_obj_by_metadata_handler)

    def _register_config_req_handlers(self):
        # Read handlers
        get_auth_rule_handler = GetAuthRuleHandler(database_manager=self.node.db_manager,
                                                   write_req_validator=self.node.write_req_validator)
        # Write handlers
        auth_rule_handler = AuthRuleHandler(database_manager=self.node.db_manager,
                                            write_req_validator=self.node.write_req_validator)
        auth_rule_handler_1_9_1 = AuthRuleHandler191(database_manager=self.node.db_manager,
                                                     write_req_validator=self.node.write_req_validator)
        auth_rules_handler = AuthRulesHandler(database_manager=self.node.db_manager,
                                              write_req_validator=self.node.write_req_validator)
        pool_config_handler = PoolConfigHandler(database_manager=self.node.db_manager,
                                                write_req_validator=self.node.write_req_validator,
                                                pool_config=self.node.poolCfg)
        pool_upgrade_handler = PoolUpgradeHandler(database_manager=self.node.db_manager,
                                                  upgrader=self.node.upgrader,
                                                  write_req_validator=self.node.write_req_validator,
                                                  pool_manager=self.node.poolManager)
        taa_aml_handler = TxnAuthorAgreementAmlHandler(database_manager=self.node.db_manager,
                                                       write_req_validator=self.node.write_req_validator)
        taa_handler = TxnAuthorAgreementHandler(database_manager=self.node.db_manager,
                                                write_req_validator=self.node.write_req_validator)
        taa_handler_v1 = TxnAuthorAgreementHandlerV1(database_manager=self.node.db_manager,
                                                     write_req_validator=self.node.write_req_validator)

        taa_disable_handler = TxnAuthorAgreementDisableHandler(database_manager=self.node.db_manager,
                                                               write_req_validator=self.node.write_req_validator)

        get_taa_aml_handler = GetTxnAuthorAgreementAmlHandler(database_manager=self.node.db_manager)
        get_taa_handler = GetTxnAuthorAgreementHandler(database_manager=self.node.db_manager)
        node_upgrade_handler = NodeUpgradeHandler(database_manager=self.node.db_manager)
        ledgers_freeze_handler = LedgersFreezeHandler(database_manager=self.node.db_manager,
                                                      write_req_validator=self.node.write_req_validator)
        get_frozen_ledgers_handler = GetFrozenLedgersHandler(database_manager=self.node.db_manager)
        # Register write handlers
        self.node.write_manager.register_req_handler(auth_rule_handler)
        self.node.write_manager.register_req_handler(auth_rules_handler)
        self.node.write_manager.register_req_handler(pool_config_handler)
        self.node.write_manager.register_req_handler(pool_upgrade_handler)
        self.node.write_manager.register_req_handler(taa_aml_handler)
        self.node.write_manager.register_req_handler(taa_handler)
        self.node.write_manager.register_req_handler(taa_disable_handler)
        self.node.write_manager.register_req_handler(node_upgrade_handler)
        self.node.write_manager.register_req_handler(ledgers_freeze_handler)
        # Register read handlers
        self.node.read_manager.register_req_handler(get_auth_rule_handler)
        self.node.read_manager.register_req_handler(get_taa_aml_handler)
        self.node.read_manager.register_req_handler(get_taa_handler)
        self.node.read_manager.register_req_handler(get_frozen_ledgers_handler)
        # Register write handlers for a version
        self.node.write_manager.register_req_handler_with_version(auth_rule_handler_1_9_1,
                                                                  version="1.9.1")
        self.node.write_manager.register_req_handler_with_version(taa_handler_v1,
                                                                  version="1")

    def _register_action_req_handlers(self):
        # Action handlers
        pool_restart_handler = PoolRestartHandler(database_manager=self.node.db_manager,
                                                  write_req_validator=self.node.write_req_validator,
                                                  restarter=self.node.restarter)
        validator_info_handler = ValidatorInfoHandler(database_manager=self.node.db_manager,
                                                      write_req_validator=self.node.write_req_validator,
                                                      info_tool=self.node._info_tool)
        # Register action handlers
        self.node.action_manager.register_action_handler(pool_restart_handler)
        self.node.action_manager.register_action_handler(validator_info_handler)

    def _register_domain_batch_handlers(self):
        super()._register_domain_batch_handlers()
        self.register_idr_cache_batch_handler()

    def _register_config_batch_handlers(self):
        config_batch_handler = ConfigBatchHandler(database_manager=self.node.db_manager,
                                                  upgrader=self.node.upgrader,
                                                  pool_config=self.node.poolCfg)
        self.node.write_manager.register_batch_handler(config_batch_handler)

    def register_idr_cache_nym_handler(self):
        idr_cache_nym_handler = IdrCacheNymHandler(database_manager=self.node.db_manager)
        self.node.write_manager.register_req_handler(idr_cache_nym_handler)

    def register_idr_cache_batch_handler(self):
        idr_cache_batch_handler = IdrCacheBatchHandler(database_manager=self.node.db_manager)
        self.node.write_manager.register_batch_handler(idr_cache_batch_handler)

    def init_pool_config(self):
        return PoolConfig(self.node.configLedger)

    def init_upgrader(self):
        return Upgrader(self.node.id,
                        self.node.name,
                        self.node.dataLocation,
                        self.node.config,
                        self.node.configLedger,
                        actionFailedCallback=self.node.postConfigLedgerCaughtUp,
                        action_start_callback=self.node.notify_upgrade_start)

    def init_restarter(self):
        return Restarter(self.node.id,
                         self.node.name,
                         self.node.dataLocation,
                         self.node.config)

    def _init_common_managers(self):
        super()._init_common_managers()
        self.node.upgrader = self.init_upgrader()
        self.node.restarter = self.init_restarter()
        self.node.poolCfg = self.init_pool_config()

    def _init_write_request_validator(self):
        constraint_serializer = ConstraintsSerializer(domain_state_serializer)
        config_state = self.node.states[CONFIG_LEDGER_ID]
        self.node.write_req_validator = WriteRequestValidator(config=self.node.config,
                                                              auth_map=auth_map,
                                                              cache=self.node.idrCache,
                                                              config_state=config_state,
                                                              state_serializer=constraint_serializer,
                                                              metrics=self.node.metrics)



from plenum.test.test_node import ensure_node_disconnected, getNonPrimaryReplicas
from indy_node.test.helper import sdk_add_raw_attribute


def test_n_minus_f_pool_processes_attrib(looper, nodeSet,
                                         sdk_pool_handle,
                                         sdk_wallet_steward):
    """
    The pool N-f nodes should be able to process ATTRIB txn.
    https://jira.hyperledger.org/browse/INDY-245
    """
    make_pool_n_minus_f_nodes(looper, nodeSet)

    sdk_add_raw_attribute(looper, sdk_pool_handle, sdk_wallet_steward, 'foo', 'bar')


def make_pool_n_minus_f_nodes(looper, nodeSet):
    non_primary, other_nodes = get_any_non_primary_and_others(nodeSet)
    disconnect_node(looper, non_primary, other_nodes)


def get_any_non_primary_and_others(node_set):
    non_primary_node = getNonPrimaryReplicas(node_set, 0)[0].node
    other_nodes = [n for n in node_set if n != non_primary_node]
    return non_primary_node, other_nodes


def disconnect_node(looper, node, other_nodes):
    node.stop()
    looper.removeProdable(node)
    ensure_node_disconnected(looper, node, other_nodes)
    check_if_pool_n_minus_f(other_nodes)


def check_if_pool_n_minus_f(nodes):
    for node in nodes:
        min_connection = node.minimumNodes - 1  # subtract node itself
        assert len(node.nodestack.connecteds) == min_connection, \
            "the pool should have minimum (N-f) nodes connected"



import copy
import json
import random

import pytest

from indy_common.constants import REVOC_TYPE, TAG, TAG_LIMIT_SIZE
from plenum.common.constants import GENERAL_LIMIT_SIZE, REQNACK, REJECT
from plenum.common.types import OPERATION
from plenum.common.util import randomString
from plenum.test.helper import sdk_sign_request_from_dict, sdk_send_signed_requests, sdk_get_replies



@pytest.fixture(scope="module", params=['lt', 'eq', 'gt'])
def _lt_eq_gt(request):
    return request.param


@pytest.fixture(scope="module", params=[REVOC_TYPE, TAG])
def _res_field_size(request, _lt_eq_gt):
    _field = request.param
    _expected = REQNACK if _lt_eq_gt == 'gt' else REJECT
    _valid_size = TAG_LIMIT_SIZE if _field == TAG else GENERAL_LIMIT_SIZE

    if _lt_eq_gt == 'lt':
        return _expected, _field, random.randint(1, _valid_size - 1)
    if _lt_eq_gt == 'eq':
        return _expected, _field, _valid_size
    return _expected, _field, random.randint(_valid_size + 1, 2 * _valid_size)


@pytest.fixture(scope="module")
def revoc_def_req(looper,
                  sdk_wallet_steward,
                  build_revoc_def_by_default,
                  _res_field_size):
    _expected, _field, _size = _res_field_size
    _req = copy.deepcopy(build_revoc_def_by_default)

    _req[OPERATION][_field] = randomString(_size)

    return _expected, sdk_sign_request_from_dict(looper, sdk_wallet_steward, _req['operation'])


def test_revoc_def_static_validation_on_field_size(revoc_def_req,
                                                   looper,
                                                   txnPoolNodeSet,
                                                   sdk_pool_handle,
                                                   sdk_wallet_steward):
    _expected, _req = revoc_def_req
    results = sdk_send_signed_requests(sdk_pool_handle, [json.dumps(_req)])
    _reply = sdk_get_replies(looper, results)[0][1]
    assert _expected == _reply['op']



import pytest

from indy_common.authorize.auth_constraints import AuthConstraintForbidden
from indy_common.constants import SCHEMA_ATTR_NAMES
from indy_common.req_utils import get_write_schema_name, get_write_schema_version, get_txn_schema_attr_names
from indy_node.server.request_handlers.domain_req_handlers.schema_handler import SchemaHandler
from indy_node.test.request_handlers.helper import add_to_idr
from plenum.common.constants import TRUSTEE
from plenum.common.exceptions import UnauthorizedClientRequest
from plenum.common.txn_util import get_request_data, reqToTxn, append_txn_metadata
from plenum.server.request_handlers.utils import encode_state_value


def make_schema_exist(schema_request, schema_handler):
    identifier, req_id, operation = get_request_data(schema_request)
    schema_name = get_write_schema_name(schema_request)
    schema_version = get_write_schema_version(schema_request)
    path = SchemaHandler.make_state_path_for_schema(identifier, schema_name, schema_version)
    schema_handler.state.set(path, encode_state_value("value", "seqNo", "txnTime"))


def test_schema_dynamic_validation_failed_existing_schema(schema_request, schema_handler):
    make_schema_exist(schema_request, schema_handler)
    with pytest.raises(UnauthorizedClientRequest, match=str(AuthConstraintForbidden())):
        schema_handler.dynamic_validation(schema_request, 0)


def test_schema_dynamic_validation_failed_not_authorised(schema_request, schema_handler):
    add_to_idr(schema_handler.database_manager.idr_cache, schema_request.identifier, None)
    with pytest.raises(UnauthorizedClientRequest):
        schema_handler.dynamic_validation(schema_request, 0)


def test_schema_dynamic_validation_passes(schema_request, schema_handler):
    add_to_idr(schema_handler.database_manager.idr_cache, schema_request.identifier, TRUSTEE)
    schema_handler.dynamic_validation(schema_request, 0)


def test_update_state(schema_request, schema_handler):
    seq_no = 1
    txn_time = 1560241033
    txn = reqToTxn(schema_request)
    append_txn_metadata(txn, seq_no, txn_time)
    path, value_bytes = SchemaHandler.prepare_schema_for_state(txn)
    value = {
        SCHEMA_ATTR_NAMES: get_txn_schema_attr_names(txn)
    }

    schema_handler.update_state(txn, None, schema_request)
    assert schema_handler.get_from_state(path) == (value, seq_no, txn_time)



import pytest

from indy_node.test.pool_config.helper import sdk_ensure_pool_config_sent


def genPoolConfig(writes: bool, force: bool):
    return dict(writes=writes, force=force)


@pytest.fixture(scope='module')
def poolConfigWTFF():
    return genPoolConfig(writes=True, force=False)


@pytest.fixture(scope='module')
def poolConfigWFFF():
    return genPoolConfig(writes=False, force=False)


@pytest.fixture(scope='module')
def poolConfigWTFT():
    return genPoolConfig(writes=True, force=True)


@pytest.fixture(scope='module')
def poolConfigWFFT():
    return genPoolConfig(writes=False, force=True)


@pytest.fixture(scope="module")
def poolConfigSent(looper, nodeSet, sdk_pool_handle, sdk_wallet_trustee, poolCfg):
    sdk_ensure_pool_config_sent(looper, sdk_pool_handle, sdk_wallet_trustee, poolCfg)



from copy import deepcopy

from indy_node.test import waits
from stp_core.loop.eventually import eventually
from plenum.common.constants import VERSION

from indy_node.test.upgrade.helper import checkUpgradeScheduled, \
    sdk_ensure_upgrade_sent, lowerVersion
from indy_common.constants import REINSTALL


def test_do_upgrade_to_the_same_version_if_reinstall(looper, tconf, nodeSet,
                                                     validUpgrade, sdk_pool_handle,
                                                     sdk_wallet_trustee):
    upgr1 = deepcopy(validUpgrade)
    upgr1[VERSION] = lowerVersion(validUpgrade['version'])
    upgr1[REINSTALL] = True

    # An upgrade scheduled, it should pass
    sdk_ensure_upgrade_sent(looper, sdk_pool_handle, sdk_wallet_trustee, upgr1)
    looper.run(
        eventually(
            checkUpgradeScheduled,
            nodeSet,
            upgr1[VERSION],
            retryWait=1,
            timeout=waits.expectedUpgradeScheduled()))
